"""Test delivery report API methods."""
import logging
from datetime import datetime, timedelta
from pathlib import Path
from typing import List

from _pytest.logging import LogCaptureFixture
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI

from cg.store import Store

from cg.meta.report.mip_dna import MipDNAReportAPI

from cg.constants import REPORT_GENDER
from cg.exc import DeliveryReportError
from cg.models.mip.mip_analysis import MipAnalysis
from cg.models.report.report import DataAnalysisModel, ReportModel, CustomerModel, CaseModel
from cg.models.report.sample import SampleModel, ApplicationModel, MethodsModel, TimestampModel
from cg.store.models import Analysis, FamilySample, Family
from tests.meta.report.helper import recursive_assert
from tests.store_helpers import StoreHelpers


def test_create_delivery_report(report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family):
    """Tests the creation of the rendered delivery report."""

    # GIVEN a pre-built case

    # WHEN extracting and rendering the report data
    delivery_report: str = report_api_mip_dna.create_delivery_report(
        case_id=case_mip_dna.internal_id,
        analysis_date=case_mip_dna.analyses[0].started_at,
        force_report=False,
    )

    # THEN check if the delivery report has been created
    assert len(delivery_report) > 0


def test_create_delivery_report_file(
    report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family, tmp_path: Path
):
    """Tests file generation containing the delivery report data."""

    # GIVEN a pre-built case

    # WHEN creating the report file
    created_report_file: Path = report_api_mip_dna.create_delivery_report_file(
        case_id=case_mip_dna.internal_id,
        directory=tmp_path,
        analysis_date=case_mip_dna.analyses[0].started_at,
        force_report=False,
    )

    # THEN check if a html report has been created and saved
    assert created_report_file.is_file()
    assert created_report_file.exists()


def test_render_delivery_report(report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family):
    """Tests delivery report rendering."""

    # GIVEN a generated report
    report_data: ReportModel = report_api_mip_dna.get_report_data(
        case_mip_dna.internal_id, case_mip_dna.analyses[0].started_at
    )

    # WHEN rendering the report
    rendered_report: str = report_api_mip_dna.render_delivery_report(report_data.dict())

    # THEN validate rendered report
    assert len(rendered_report) > 0
    assert "html" in rendered_report


def test_get_validated_report_data(report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family):
    """Tests report data retrieval."""

    # GIVEN a valid case
    assert case_mip_dna
    assert case_mip_dna.links
    assert case_mip_dna.analyses

    # WHEN collecting the delivery data
    report_data: ReportModel = report_api_mip_dna.get_report_data(
        case_mip_dna.internal_id, case_mip_dna.analyses[0].started_at
    )

    # THEN check collection of the nested report data and that the required fields are not empty
    report_data: ReportModel = report_api_mip_dna.validate_report_fields(
        case_mip_dna.internal_id, report_data, force_report=False
    )
    recursive_assert(report_data.dict())


def test_validate_report_empty_fields(
    report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family, caplog: LogCaptureFixture
):
    """Tests the validations of allowed empty report fields."""
    caplog.set_level(logging.INFO)

    # GIVEN a delivery report
    report_data: ReportModel = report_api_mip_dna.get_report_data(
        case_mip_dna.internal_id, case_mip_dna.analyses[0].started_at
    )

    # WHEN the report has some allowed empty fields
    report_data.version = None
    report_data.customer.id = None
    report_data.case.samples[0].methods.library_prep = None

    # THEN check if the empty fields are identified
    report_data: ReportModel = report_api_mip_dna.validate_report_fields(
        case_mip_dna.internal_id, report_data, force_report=False
    )
    assert report_data
    assert "version" in caplog.text
    assert "id" in caplog.text
    assert "library_prep" in caplog.text


def test_validate_report_missing_fields(
    report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family, caplog: LogCaptureFixture
):
    """Tests the validations of empty required report fields."""

    # GIVEN a delivery report
    report_data: ReportModel = report_api_mip_dna.get_report_data(
        case_mip_dna.internal_id, case_mip_dna.analyses[0].started_at
    )

    # WHEN the report contains some required empty fields
    report_data.accredited = None
    report_data.case.samples[0].metadata.million_read_pairs = None
    report_data.case.samples[1].metadata.duplicates = None

    # THEN test that the DeliveryReportError is raised when the report generation is not forced
    try:
        report_api_mip_dna.validate_report_fields(
            case_mip_dna.internal_id, report_data, force_report=False
        )
    except DeliveryReportError:
        assert "accredited" in caplog.text
        assert "duplicates" in caplog.text
        assert "accredited" in caplog.text


def test_get_validated_report_data_external_sample(
    report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family
):
    """Tests report data retrieval."""

    # GIVEN a delivery report with external sample data
    report_data: ReportModel = report_api_mip_dna.get_report_data(
        case_mip_dna.internal_id, case_mip_dna.analyses[0].started_at
    )
    report_data.case.samples[0].timestamps.received_at = None
    report_data.case.samples[0].application.external = True

    # WHEN validating report fields
    report_data: ReportModel = report_api_mip_dna.validate_report_fields(
        case_mip_dna.internal_id, report_data, force_report=False
    )

    # THEN the validation should have been completed successfully
    assert report_data


def test_get_customer_data(report_api_mip_dna: MipDNAReportAPI, case_mip_dna: Family):
    """Checks that the retrieved customer data is the expected one."""

    # GIVEN a pre-built case

    # GIVEN an expected output
    expected_customer: dict = {
        "name": "Production",
        "id": "cust000",
        "invoice_address": "Test street",
        "scout_access": True,
    }

    # WHEN retrieving the customers data
    customer_data: CustomerModel = report_api_mip_dna.get_customer_data(case_mip_dna)

    # THEN check if the retrieved customer data corresponds to the expected one
    assert customer_data == expected_customer


def test_get_report_version_version(
    report_api_mip_dna: MipDNAReportAPI,
    store: Store,
    helpers: StoreHelpers,
    timestamp_yesterday: datetime,
):
    """Validates the extracted report versions of two analyses."""

    # GIVEN a specific set of analyses
    last_analysis: Analysis = helpers.add_analysis(store, completed_at=datetime.now())
    first_analysis: Analysis = helpers.add_analysis(
        store, last_analysis.family, completed_at=timestamp_yesterday
    )

    # WHEN retrieving the version
    last_analysis_version: int = report_api_mip_dna.get_report_version(last_analysis)
    first_analysis_version: int = report_api_mip_dna.get_report_version(first_analysis)

    # THEN check if the versions match
    assert last_analysis_version == 2
    assert first_analysis_version == 1


def test_get_case_data(
    report_api_mip_dna: MipDNAReportAPI,
    mip_analysis_api: MipDNAAnalysisAPI,
    case_mip_dna: Family,
    family_name: str,
):
    """Tests the extracted case data."""

    # GIVEN a pre-built case

    # GIVEN a mip analysis mock metadata
    mip_metadata: MipAnalysis = mip_analysis_api.get_latest_metadata(case_mip_dna.internal_id)

    # WHEN retrieving case specific information
    case_data: CaseModel = report_api_mip_dna.get_case_data(
        case_mip_dna, case_mip_dna.analyses[0], mip_metadata
    )

    # THEN check if the case data is the expected one
    assert case_data.name == family_name
    assert case_data.data_analysis
    assert case_data.samples
    assert case_data.applications


def test_get_samples_data(
    report_api_mip_dna: MipDNAReportAPI,
    mip_analysis_api: MipDNAAnalysisAPI,
    case_mip_dna: Family,
    case_samples_data: List[FamilySample],
    lims_samples: List[dict],
):
    """Validates the retrieved sample data."""

    # GIVEN a pre-built case

    # GIVEN an expected output
    expected_lims_data: dict = lims_samples[0]
    expected_sample_data: FamilySample = case_samples_data[0]

    # GIVEN a mip analysis mock metadata
    mip_metadata: MipAnalysis = mip_analysis_api.get_latest_metadata(case_mip_dna.internal_id)

    # WHEN extracting the samples of a specific case
    samples_data: SampleModel = report_api_mip_dna.get_samples_data(case_mip_dna, mip_metadata)[0]

    # THEN assert that the retrieved sample is the expected one
    assert samples_data.name == str(expected_sample_data.sample.name)
    assert samples_data.id == str(expected_sample_data.sample.internal_id)
    assert samples_data.ticket == str(expected_sample_data.sample.original_ticket)
    assert samples_data.status == str(expected_sample_data.status)
    assert samples_data.gender == REPORT_GENDER.get(str(expected_sample_data.sample.sex))
    assert samples_data.source == str(expected_lims_data.get("source"))
    assert samples_data.tumour == "Nej"
    assert samples_data.application
    assert samples_data.methods
    assert samples_data.metadata
    assert samples_data.timestamps


def test_get_lims_sample(
    report_api_mip_dna: MipDNAReportAPI,
    case_samples_data: List[FamilySample],
    lims_samples: List[dict],
):
    """Tests lims data extraction."""

    # GIVEN a family samples instance

    # GIVEN an expected output
    expected_lims_data: dict = lims_samples[0]

    # WHEN getting the sample data from lims
    lims_data: dict = report_api_mip_dna.get_lims_sample(case_samples_data[0].sample.internal_id)

    # THEN check if the extracted lims information match the expected one
    assert lims_data == expected_lims_data


def test_get_sample_application_data(
    report_api_mip_dna: MipDNAReportAPI,
    case_samples_data: List[FamilySample],
    lims_samples: List[dict],
):
    """Tests sample application data extraction."""

    # GIVEN a lims sample instance

    # GIVEN the expected application data
    expected_application_data: dict = case_samples_data[0].sample.to_dict().get("application")

    # WHEN retrieving application data from status DB
    application_data: ApplicationModel = report_api_mip_dna.get_sample_application_data(
        lims_samples[0]
    )

    # THEN verify that the application data corresponds to what is expected
    assert application_data.tag == str(expected_application_data.get("tag"))
    assert application_data.version == str(lims_samples[0].get("application_version"))
    assert application_data.prep_category == str(expected_application_data.get("prep_category"))
    assert application_data.description == str(expected_application_data.get("description"))
    assert application_data.limitations == str(expected_application_data.get("limitations"))
    assert application_data.accredited == expected_application_data.get("is_accredited")


def test_get_unique_applications(
    report_api_mip_dna: MipDNAReportAPI, mip_analysis_api: MipDNAAnalysisAPI, case_mip_dna: Family
):
    """Tests unique applications filtering."""

    # GIVEN a list of samples sharing the same application
    mip_metadata: MipAnalysis = mip_analysis_api.get_latest_metadata(case_mip_dna.internal_id)
    samples: List[SampleModel] = report_api_mip_dna.get_samples_data(case_mip_dna, mip_metadata)

    # WHEN calling the application filtering function
    unique_applications: List[ApplicationModel] = report_api_mip_dna.get_unique_applications(
        samples
    )

    # THEN check that the same sample applications were filtered out
    assert len(unique_applications) == 1


def test_get_sample_methods_data(
    report_api_mip_dna: MipDNAReportAPI, case_samples_data: List[FamilySample]
):
    """Tests sample methods retrieval from lims."""

    # GIVEN a sample ID
    sample_id: str = case_samples_data[0].sample.to_dict().get("internal_id")

    # GIVEN an expected output
    expected_sample_methods = {
        "library_prep": "CG002 - End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)",
        "sequencing": "CG002 - Cluster Generation (HiSeq X)",
    }

    # WHEN calling the method extraction function
    sample_methods: MethodsModel = report_api_mip_dna.get_sample_methods_data(sample_id)

    # THEN check the agreement between expected and extracted values
    assert sample_methods == expected_sample_methods


def test_get_case_analysis_data(
    report_api_mip_dna: MipDNAReportAPI, mip_analysis_api: MipDNAAnalysisAPI, case_mip_dna: Family
):
    """Tests data analysis parameters retrieval."""

    # GIVEN a pre-built case

    # GIVEN a mip analysis mock metadata
    mip_metadata: MipAnalysis = mip_analysis_api.get_latest_metadata(case_mip_dna.internal_id)

    # WHEN retrieving analysis information
    case_analysis_data: DataAnalysisModel = report_api_mip_dna.get_case_analysis_data(
        case_mip_dna, case_mip_dna.analyses[0], mip_metadata
    )

    # THEN check if the retrieved analysis data is correct
    assert case_analysis_data.pipeline == "mip-dna"
    assert case_analysis_data.panels == "IEM, EP"
    assert case_analysis_data.scout_files


def test_get_sample_timestamp_data(
    report_api_mip_dna: MipDNAReportAPI,
    case_samples_data: List[FamilySample],
    timestamp_yesterday: datetime,
):
    """Checks that the sample timestamp information is correctly retrieved from StatusDB."""

    # GIVEN a mock sample data

    # GIVEN an expected output
    expected_case_samples_data = {
        "ordered_at": str((datetime.now() - timedelta(days=3)).date()),
        "received_at": str((datetime.now() - timedelta(days=2)).date()),
        "prepared_at": str(timestamp_yesterday.date()),
        "sequenced_at": str(timestamp_yesterday.date()),
    }

    # WHEN extracting the timestamp data associated to a specific sample
    sample_timestamp_data: TimestampModel = report_api_mip_dna.get_sample_timestamp_data(
        case_samples_data[0].sample
    )

    # THEN check if the dates are correctly retrieved
    assert sample_timestamp_data == expected_case_samples_data



import logging
from pathlib import Path
from typing import Dict, List, Optional, Set, Hashable, Iterable

from pydantic.v1 import constr, BaseModel

from cg.constants import DataDelivery
from cg.exc import OrderFormError
from cg.models.orders.order import OrderType
from cg.models.orders.orderform_schema import OrderCase, Orderform, OrderPool
from cg.models.orders.sample_base import OrderSample
from cg.store.models import Customer

LOG = logging.getLogger(__name__)


class OrderformParser(BaseModel):
    """Class to parse orderforms"""

    samples: List[OrderSample] = []
    project_type: Optional[OrderType] = None
    delivery_type: Optional[DataDelivery] = None
    customer_id: constr(
        min_length=1, max_length=Customer.internal_id.property.columns[0].type.length
    ) = None
    order_comment: Optional[str] = None
    order_name: Optional[str] = None

    def parse_orderform(self, orderform_file: Path) -> None:
        """Parse the orderform information"""
        raise NotImplementedError

    def group_cases(self) -> Dict[str, List[OrderSample]]:
        """Group samples in cases."""
        LOG.info("Group samples under respective case")
        cases = {}
        for sample in self.samples:
            case_id = sample.family_name
            if not case_id:
                continue
            if case_id not in cases:
                cases[case_id] = []
            cases[case_id].append(sample)
        if cases:
            LOG.info("Found cases %s", ", ".join(cases.keys()))
        else:
            LOG.info("Could not find any cases")
        return cases

    def get_pools(self) -> List[OrderPool]:
        """create pools from samples

        Check that all samples from one pool has the same application
        """
        pools: Dict[str, OrderPool] = {}
        sample: OrderSample
        for sample in self.samples:
            pool_name = sample.pool
            if not pool_name:
                continue
            application = sample.application
            if pool_name not in pools:
                pools[pool_name] = OrderPool(
                    name=pool_name,
                    data_analysis=sample.data_analysis,
                    data_delivery=sample.data_delivery,
                    application=application,
                    samples=[sample],
                )
                continue
            # each pool must only have one application type
            pool = pools[pool_name]
            if str(pool.application) != application:
                raise OrderFormError(
                    f"different applications in pool: {pool_name} - {[pools[pool_name].application, application]}"
                )
            pool.samples.append(sample)

        return [pools[pool_name] for pool_name in pools]

    @staticmethod
    def _get_single_value(
        items_id: str, items: Iterable, attr: str, default_value: Hashable = None
    ) -> Hashable:
        """
        Get single value from a bunch of items, will raise if value not same on all items
        @param items_id:        id of items group (e.g. case-id)
        @param items:           some items (e.g. samples)
        @param attr:            one attribute of item containing a value (e.g. synopsis)
        @param default_value:   default value to use if item has an empty value of attr
        @return:                a value with the attribute of item in items
        """
        values: Set[Hashable] = set(getattr(item, attr, default_value) for item in items)
        if len(values) > 1:
            raise OrderFormError(f"multiple values [{values}] for '{attr}' for '{items_id}'")

        return values.pop()

    @staticmethod
    def _get_single_set(items_id: str, items: Iterable, attr: str) -> Set[Hashable]:
        """
        Get single value (set) from a bunch of items, will raise if value not same on all items
        @param items_id: id of items group (e.g. case-id)
        @param items:    some items (e.g. samples)
        @param attr:     one attribute of item containing a list (e.g. panels)
        @return:         a set with the attribute of item in items
        """
        values: Set[Hashable] = set()
        for item_idx, item in enumerate(items):
            if item_idx == 0:
                values = set(getattr(item, attr)) if getattr(item, attr) else set()
            elif values != set(getattr(item, attr)) if getattr(item, attr) else set():
                raise OrderFormError(f"multiple values [{values}] for '{attr}' for '{items_id}'")
        return values

    @staticmethod
    def expand_case(case_id: str, case_samples: List[OrderSample]) -> OrderCase:
        """Fill-in information about case."""

        priority: Hashable[str] = OrderformParser._get_single_value(
            items_id=case_id, items=case_samples, attr="priority"
        )
        synopsis: Hashable[str] = OrderformParser._get_single_value(
            items_id=case_id, items=case_samples, attr="synopsis"
        )
        cohorts: Set[Hashable[str]] = OrderformParser._get_single_set(
            items_id=case_id, items=case_samples, attr="cohorts"
        )
        panels: Set[Hashable[str]] = OrderformParser._get_single_set(
            items_id=case_id, items=case_samples, attr="panels"
        )

        return OrderCase(
            cohorts=list(cohorts),
            name=case_id,
            samples=case_samples,
            priority=priority,
            panels=list(panels),
            synopsis=synopsis,
        )

    def generate_orderform(self) -> Orderform:
        """Generate an orderform"""
        cases_map: Dict[str, List[OrderSample]] = self.group_cases()
        case_objs: List[OrderCase] = []
        for case_id in cases_map:
            case_objs.append(self.expand_case(case_id=case_id, case_samples=cases_map[case_id]))
        return Orderform(
            comment=self.order_comment,
            samples=self.samples,
            cases=case_objs,
            name=self.order_name,
            customer=self.customer_id,
            delivery_type=self.delivery_type,
            project_type=self.project_type,
            pools=self.get_pools(),
        )

    def __repr__(self):
        return (
            f"OrderformParser(project_type={self.project_type},delivery_type={self.delivery_type},customer_id="
            f"{self.customer_id},order_name={self.order_name})"
        )



import logging
from pathlib import Path
from typing import List

import click
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.sample_sheet.create import create_sample_sheet
from cg.apps.demultiplex.sample_sheet.models import FlowCellSample
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_sample_sheet_from_file
from cg.apps.lims.sample_sheet import get_flow_cell_samples
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import OPTION_BCL_CONVERTER
from cg.exc import FlowCellError
from cg.io.controller import WriteFile, WriteStream
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from pydantic.v1 import ValidationError

LOG = logging.getLogger(__name__)


@click.group(name="samplesheet")
def sample_sheet_commands():
    """Command group for the sample sheet commands."""


@sample_sheet_commands.command(name="validate")
@click.argument("flow-cell-name")
@click.argument("sheet", type=click.Path(exists=True, dir_okay=False))
@OPTION_BCL_CONVERTER
@click.pass_obj
def validate_sample_sheet(
    context: CGConfig,
    flow_cell_name: str,
    bcl_converter: str,
    sheet: click.Path,
):
    """Validate a sample sheet.
    flow-cell-name is the flow cell run directory name, e.g. '201203_A00689_0200_AHVKJCDRXX'
    """

    flow_cell_path: Path = Path(context.demultiplex_api.run_dir, flow_cell_name)
    flow_cell: FlowCellDirectoryData = FlowCellDirectoryData(
        flow_cell_path=flow_cell_path, bcl_converter=bcl_converter
    )
    LOG.info(
        f"Validating {sheet} as a {flow_cell.sequencer_type} {bcl_converter} sample sheet",
    )
    sheet: Path = Path(str(sheet))
    try:
        get_sample_sheet_from_file(infile=sheet, flow_cell_sample_type=flow_cell.sample_type)
    except ValidationError as error:
        LOG.warning(error)
        raise click.Abort from error
    LOG.info("Sample sheet passed validation")


@sample_sheet_commands.command(name="create")
@click.argument("flow-cell-name")
@OPTION_BCL_CONVERTER
@click.option("--dry-run", is_flag=True)
@click.option("--force", is_flag=True, help="Skips the validation of the sample sheet")
@click.pass_obj
def create_sheet(
    context: CGConfig, flow_cell_name: str, bcl_converter: str, dry_run: bool, force: bool = False
):
    """Command to create a sample sheet.
    flow-cell-name is the flow cell run directory name, e.g. '201203_A00689_0200_AHVKJCDRXX'

    Search the flow cell in the directory specified in config.
    """

    LOG.info(f"Creating sample sheet for flow cell {flow_cell_name}")
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    flow_cell_path: Path = Path(demultiplex_api.run_dir, flow_cell_name)
    if not flow_cell_path.exists():
        LOG.warning(f"Could not find flow cell {flow_cell_path}")
        raise click.Abort
    try:
        flow_cell = FlowCellDirectoryData(
            flow_cell_path=flow_cell_path, bcl_converter=bcl_converter
        )
    except FlowCellError as error:
        raise click.Abort from error
    lims_samples: List[FlowCellSample] = list(
        get_flow_cell_samples(
            lims=context.lims_api,
            flow_cell_id=flow_cell.id,
            flow_cell_sample_type=flow_cell.sample_type,
        )
    )
    if not lims_samples:
        LOG.warning(f"Could not find any samples in lims for {flow_cell.id}")
        raise click.Abort
    try:
        sample_sheet_content: List[List[str]] = create_sample_sheet(
            bcl_converter=bcl_converter, flow_cell=flow_cell, lims_samples=lims_samples, force=force
        )
    except (FileNotFoundError, FileExistsError, FlowCellError) as error:
        raise click.Abort from error

    if dry_run:
        click.echo(
            WriteStream.write_stream_from_content(
                file_format=FileFormat.CSV, content=sample_sheet_content
            )
        )
        return
    LOG.info(f"Writing sample sheet to {flow_cell.sample_sheet_path.resolve()}")
    WriteFile.write_file_from_content(
        content=sample_sheet_content,
        file_format=FileFormat.CSV,
        file_path=flow_cell.sample_sheet_path,
    )


@sample_sheet_commands.command(name="create-all")
@OPTION_BCL_CONVERTER
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def create_all_sheets(context: CGConfig, bcl_converter: str, dry_run: bool):
    """Command to create sample sheets for all flow cells that lack a sample sheet.

    Search flow cell directories for run parameters and create a sample sheets based on the
    information.
    """
    demux_api: DemultiplexingAPI = context.demultiplex_api
    flow_cells: Path = demux_api.run_dir
    for sub_dir in flow_cells.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.info(f"Found directory {sub_dir}")
        try:
            flow_cell = FlowCellDirectoryData(flow_cell_path=sub_dir, bcl_converter=bcl_converter)
        except FlowCellError:
            continue
        if flow_cell.sample_sheet_exists():
            LOG.info("Sample sheet already exists")
            continue
        LOG.info(f"Creating sample sheet for flow cell {flow_cell.id}")
        lims_samples: List[FlowCellSample] = list(
            get_flow_cell_samples(
                lims=context.lims_api,
                flow_cell_id=flow_cell.id,
                flow_cell_sample_type=flow_cell.sample_type,
            )
        )
        if not lims_samples:
            LOG.warning(f"Could not find any samples in lims for {flow_cell.id}")
            continue

        try:
            sample_sheet_content: List[List[str]] = create_sample_sheet(
                flow_cell=flow_cell, lims_samples=lims_samples, bcl_converter=bcl_converter
            )
        except (FileNotFoundError, FileExistsError, ValidationError, FlowCellError):
            continue

        if dry_run:
            click.echo(
                WriteStream.write_stream_from_content(
                    file_format=FileFormat.CSV, content=sample_sheet_content
                )
            )
            continue
        LOG.info(f"Writing sample sheet to {flow_cell.sample_sheet_path.resolve()}")
        WriteFile.write_file_from_content(
            content=sample_sheet_content,
            file_format=FileFormat.CSV,
            file_path=flow_cell.sample_sheet_path,
        )



import logging
import pytest

from pathlib import Path
from typing import List, Optional

from cg.apps.cgstats.db import models
from cg.apps.cgstats.stats import StatsAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import DeleteDemuxError
from cg.meta.demultiplex.delete_demultiplex_api import DeleteDemuxAPI
from cg.models.cg_config import CGConfig
from cg.store.api import Store
from cg.store.models import Sample, Flowcell
from tests.store_helpers import StoreHelpers


def test_initiate_delete_demux_api(
    caplog,
    cg_context: CGConfig,
    flow_cell_name: str,
    tmp_flow_cell_run_base_path: Path,
    bcl2fastq_flow_cell_id: str,
):
    """Test to initialize the DeleteDemuxAPI"""

    caplog.set_level(logging.DEBUG)

    # GIVEN a correct config
    config = cg_context
    config.demultiplex_api.run_dir = tmp_flow_cell_run_base_path
    config.demultiplex_api.out_dir = tmp_flow_cell_run_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # WHEN initializing the DeleteDemuxAPI
    DeleteDemuxAPI(
        config=config,
        flow_cell_name=bcl2fastq_flow_cell_id,
        dry_run=True,
    )

    # THEN the API should be correctly initialized
    assert "DeleteDemuxAPI: API initiated" in caplog.text


def test_get_presence_status_status_db(
    caplog,
    helpers: StoreHelpers,
    delete_demultiplex_api: DeleteDemuxAPI,
    bcl2fastq_flow_cell_id: str,
):
    """Test to see if the presence of a flow cell is detected in status-db."""
    caplog.set_level(logging.INFO)
    # GIVEN DeleteDemuxAPI objects, one with amd one without a flowcell in status-db
    delete_demux_api: DeleteDemuxAPI = delete_demultiplex_api

    # WHEN the flowcell name is parsed and fetching the presence of a flowcell in either context
    empty_presence: bool = delete_demux_api.status_db_presence

    # THEN there should be an appropriate presence in both cases
    assert not empty_presence

    # WHEN adding a flowcell into the statusdb and checking its updated presence
    helpers.add_flowcell(
        store=delete_demux_api.status_db,
        flow_cell_name=bcl2fastq_flow_cell_id,
        sequencer_type="novaseq",
    )
    populated_presence: bool = delete_demux_api.status_db_presence

    # THEN the presence should be updated
    assert populated_presence


def test_set_dry_run_delete_demux_api(
    caplog,
    cg_context: CGConfig,
    bcl2fastq_flow_cell_id: str,
    stats_api: StatsAPI,
    tmp_flow_cell_run_base_path: Path,
):
    """Test to test function to set the API to run in dry run mode"""

    caplog.set_level(logging.DEBUG)
    cg_context.cg_stats_api_ = stats_api
    cg_context.demultiplex_api.run_dir = tmp_flow_cell_run_base_path
    cg_context.demultiplex_api.out_dir = tmp_flow_cell_run_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # WHEN setting the dry_run mode on a DeleteDemuxAPI
    delete_demultiplex_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=cg_context,
        dry_run=True,
        flow_cell_name=bcl2fastq_flow_cell_id,
    )

    # THEN the dry run parameter should be set to True and it should be logged
    assert delete_demultiplex_api.dry_run
    assert f"DeleteDemuxAPI: Setting dry run mode to True" in caplog.text


def test_no_active_samples_on_flow_cell(
    populated_delete_demultiplex_api: DeleteDemuxAPI, bcl2fastq_flow_cell_id: str
):
    """Test if the function to find no active samples works correctly."""

    # GIVEN a flow cell with no active samples related to it
    store: Store = populated_delete_demultiplex_api.status_db
    flow_cell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    samples_on_flow_cell: List[Sample] = flow_cell.samples

    assert samples_on_flow_cell
    for sample in samples_on_flow_cell:
        active: bool = store.has_active_cases_for_sample(internal_id=sample.internal_id)
        assert not active

    # WHEN checking for active samples on flowcell
    populated_delete_demultiplex_api._set_samples_on_flow_cell()
    active_samples_on_flow_cell: Optional[
        List[str]
    ] = populated_delete_demultiplex_api.active_samples_on_flow_cell()

    # THEN the no samples on the flowcell should be found active
    assert not active_samples_on_flow_cell


def test_active_samples_on_flow_cell(
    active_flow_cell_store: Store,
    bcl2fastq_flow_cell_id: str,
    active_delete_demultiplex_api: DeleteDemuxAPI,
):
    """Test if the function to find active samples works correctly."""
    # GIVEN a flow cell with active samples related to it
    store: Store = active_flow_cell_store
    flow_cell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    samples_on_flow_cell: List[Sample] = flow_cell.samples

    assert samples_on_flow_cell
    for sample in samples_on_flow_cell:
        active: bool = store.has_active_cases_for_sample(internal_id=sample.internal_id)
        assert active

    # WHEN checking for active samples on flowcell
    active_delete_demultiplex_api._set_samples_on_flow_cell()
    active_samples_on_flow_cell: Optional[
        List[str]
    ] = active_delete_demultiplex_api.active_samples_on_flow_cell()

    # THEN there should be active samples found
    assert any(sample.internal_id in active_samples_on_flow_cell for sample in samples_on_flow_cell)


def test_check_active_sample(active_delete_demultiplex_api: DeleteDemuxAPI):
    """Test that proper exception is raised when active samples are identified"""

    # GIVEN a DeleteDemuxAPI and a store with active samples related to it

    delete_demux_api: DeleteDemuxAPI = active_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)

    # WHEN checking if there are active samples on flowcell to be deleted

    with pytest.raises(DeleteDemuxError):
        # THEN the proper error should be raised
        delete_demux_api.check_active_samples()


def test_delete_flow_cell_housekeeper_only_sample_level(
    caplog,
    cg_context: CGConfig,
    tmp_flow_cell_run_base_path: Path,
    bcl2fastq_flow_cell_id: str,
    populated_flow_cell_store: Store,
    sample_level_housekeeper_api: HousekeeperAPI,
):
    """Test removing fastqs from Housekeeper when there are only files on sample level
    (not on flow cell name).
    """

    caplog.set_level(logging.INFO)

    cg_context.demultiplex_api.run_dir = tmp_flow_cell_run_base_path
    cg_context.demultiplex_api.out_dir = tmp_flow_cell_run_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # GIVEN a DeleteDemuxAPI with a HousekeeperAPI with no files with flow cell name as a tag

    delete_demultiplex_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=cg_context,
        dry_run=False,
        flow_cell_name=bcl2fastq_flow_cell_id,
    )
    delete_demultiplex_api.housekeeper_api = sample_level_housekeeper_api
    delete_demultiplex_api._set_samples_on_flow_cell()

    # WHEN wiping files in Housekeeper

    delete_demultiplex_api.delete_flow_cell_housekeeper()

    # THEN you should be notified that there are no files on flow cell names

    assert (
        f"Housekeeper: No files found with tag: {delete_demultiplex_api.flow_cell_name}"
        in caplog.text
    )

    # THEN you should be notified that there were fastq files removed on sample level
    assert "Deleting file" in caplog.text


def test_delete_flow_cell_housekeeper_flowcell_name(
    caplog,
    cg_context: CGConfig,
    tmp_flow_cell_run_base_path: Path,
    tmp_flow_cell_demux_base_path: Path,
    flow_cell_name_housekeeper_api: HousekeeperAPI,
    bcl2fastq_flow_cell_id: str,
    populated_flow_cell_store: Store,
    tmp_sample_sheet_path: Path,
):
    """Test removing files from Housekeeper using flow cell name as a tag."""

    caplog.set_level(logging.INFO)
    cg_context.housekeeper_api_ = flow_cell_name_housekeeper_api
    cg_context.status_db_ = populated_flow_cell_store
    cg_context.demultiplex_api.run_dir = tmp_flow_cell_run_base_path
    cg_context.demultiplex_api.out_dir = tmp_flow_cell_demux_base_path
    Path(tmp_flow_cell_run_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    Path(tmp_flow_cell_demux_base_path, f"some_prefix_1100_{bcl2fastq_flow_cell_id}").mkdir(
        parents=True, exist_ok=True
    )
    # GIVEN

    sample_sheet_file: Path = tmp_sample_sheet_path

    delete_demultiplex_api: DeleteDemuxAPI = DeleteDemuxAPI(
        config=cg_context,
        dry_run=False,
        flow_cell_name=bcl2fastq_flow_cell_id,
    )
    delete_demultiplex_api.housekeeper_api = flow_cell_name_housekeeper_api
    delete_demultiplex_api._set_samples_on_flow_cell()

    # WHEN

    delete_demultiplex_api.delete_flow_cell_housekeeper()

    # THEN

    assert (
        f"Housekeeper: No files found with tag: {delete_demultiplex_api.flow_cell_name}"
        not in caplog.text
    )
    assert f"Deleted {sample_sheet_file.as_posix()} from housekeeper" in caplog.text

    # THEN you should be notified that there were fastq files removed on sample level
    assert "Deleting file" in caplog.text


def test_delete_flow_cell_statusdb(
    caplog,
    bcl2fastq_flow_cell_id: str,
    populated_delete_demultiplex_api: DeleteDemuxAPI,
    populated_delete_demux_context: CGConfig,
):
    """Test if function to remove flow cell objects from status db is working."""

    caplog.set_level(logging.INFO)

    # GIVEN a context, with a status db filled with a flow cell object
    delete_demux_api: DeleteDemuxAPI = populated_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)
    store = populated_delete_demux_context.status_db
    flow_cell: Flowcell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    assert flow_cell

    # WHEN removing the flow cell from the database
    delete_demux_api.delete_flow_cell_in_status_db()

    # THEN the user should be informed that the flow cell was removed
    assert f"StatusDB: Deleted flowcell {delete_demux_api.flow_cell_name}" in caplog.text

    # AND the flow cell should no longer exist in status db
    flow_cell: Flowcell = store.get_flow_cell_by_name(flow_cell_name=bcl2fastq_flow_cell_id)
    assert not flow_cell


def test_delete_flow_cell_hasta(
    caplog,
    populated_delete_demultiplex_api: DeleteDemuxAPI,
):
    """Test if function to remove files from the file system is working"""

    caplog.set_level(logging.INFO)
    delete_demux_api: DeleteDemuxAPI = populated_delete_demultiplex_api

    flow_cell_obj: Flowcell = delete_demux_api.status_db.get_flow_cell_by_name(
        delete_demux_api.flow_cell_name
    )
    delete_demux_api.set_dry_run(dry_run=False)

    # GIVEN an existing demultiplexing and run directory of a flow cell, with a status "ondisk"

    assert delete_demux_api.demultiplexing_out_path.exists()
    assert delete_demux_api.run_path.exists()
    assert flow_cell_obj.status == "ondisk"

    # WHEN removing said files with the DeleteDemuxAPI

    delete_demux_api.delete_flow_cell_hasta(
        demultiplexing_dir=True,
        run_dir=True,
    )

    # THEN the demultiplexing directory should be removed
    assert (
        f"DeleteDemuxAPI-Hasta: Removing flow cell demultiplexing directory: {delete_demux_api.demultiplexing_out_path}"
        in caplog.text
    )
    assert delete_demux_api.demultiplexing_out_path.exists() is False

    # THEN the run directory should be removed
    assert (
        f"DeleteDemuxAPI-Hasta: Removing flow cell run directory: {delete_demux_api.run_path}"
        in caplog.text
    )
    assert delete_demux_api.run_path.exists() is False

    # THEN the status of the flow cell in statusdb should be set to removed
    assert flow_cell_obj.status == "removed"


def test_delete_flow_cell_cgstats(
    caplog,
    populated_delete_demux_context: CGConfig,
    populated_delete_demultiplex_api: DeleteDemuxAPI,
    bcl2fastq_flow_cell_id: str,
):
    """Test if function to remove objects from cg-stats is working."""

    caplog.set_level(logging.INFO)
    delete_demux_api: DeleteDemuxAPI = populated_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)

    # GIVEN an existing object in cg-stags database

    existing_object: models.Flowcell = (
        populated_delete_demux_context.cg_stats_api.query(models.Flowcell)
        .filter(models.Flowcell.flowcellname == bcl2fastq_flow_cell_id)
        .first()
    )

    assert existing_object

    # WHEN wiping the existence of said object

    delete_demux_api.delete_flow_cell_cgstats()

    # THEN the user should be notified that the object was removed

    assert f"Removing entry {bcl2fastq_flow_cell_id} in from cgstats" in caplog.text

    # AND the object should no longer exist

    existing_object: models.Flowcell = (
        populated_delete_demux_context.cg_stats_api.query(models.Flowcell)
        .filter(models.Flowcell.flowcellname == bcl2fastq_flow_cell_id)
        .first()
    )

    assert not existing_object


def test_delete_demultiplexing_init_files(
    caplog, demultiplexing_init_files: List[Path], populated_delete_demultiplex_api: DeleteDemuxAPI
):
    """Test function to remove demultiplexing init files from the filesystem"""

    # GIVEN a list of paths to existing demultiplexing init filesystem and a initiated DeleteDemuxAPI
    delete_demux_api = populated_delete_demultiplex_api
    delete_demux_api.set_dry_run(dry_run=False)

    assert all(init_file.exists() for init_file in demultiplexing_init_files)

    # WHEN wiping the existance of mentioned files

    delete_demux_api.delete_demux_init_files()

    # THEN the files should no longer exist

    assert not any(init_file.exists() for init_file in demultiplexing_init_files)



from sqlalchemy import Column, orm, types

from .base import Model


class Flowcell(Model):
    flowcell_id = Column(types.Integer, primary_key=True)
    flowcellname = Column(types.String(255), nullable=False, unique=True)
    flowcell_pos = Column(types.Enum("A", "B"), nullable=False)
    hiseqtype = Column(types.String(255), nullable=False)
    time = Column(types.DateTime)

    demux = orm.relationship("Demux", cascade="all", backref=orm.backref("flowcell"))



"""Tests for the creation of the sample sheet"""
from typing import List
from cg.apps.demultiplex.sample_sheet.create import get_sample_sheet_creator
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreator,
    SampleSheetCreatorV1,
    SampleSheetCreatorV2,
)
from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleNovaSeqX,
    FlowCellSampleNovaSeq6000Bcl2Fastq,
)
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


def test_sample_sheet_creator_factory_novaseq_6000(
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    lims_novaseq_bcl2fastq_samples: List[FlowCellSampleNovaSeq6000Bcl2Fastq],
):
    """Test that a sample sheet creator defined with NovaSeq6000 data is V1."""
    # GIVEN a NovaSeq6000 flow cell and a list of NovaSeq6000 samples

    # WHEN defining the sample sheet creator
    sample_sheet_creator: SampleSheetCreator = get_sample_sheet_creator(
        bcl_converter=bcl2fastq_flow_cell.bcl_converter,
        flow_cell=bcl2fastq_flow_cell,
        lims_samples=lims_novaseq_bcl2fastq_samples,
        force=False,
    )

    # THEN no error is raised and the sample sheet creator is v1
    assert isinstance(sample_sheet_creator, SampleSheetCreatorV1)


def test_sample_sheet_creator_factory_novaseq_x(
    novaseq_x_flow_cell: FlowCellDirectoryData,
    lims_novaseq_x_samples: List[FlowCellSampleNovaSeqX],
):
    """Test that a sample sheet creator defined with NovaSeq6000 data is V1."""
    # GIVEN a NovaSeqX flow cell and a list of NovaSeqX samples

    # WHEN defining the sample sheet creator
    sample_sheet_creator: SampleSheetCreator = get_sample_sheet_creator(
        bcl_converter=novaseq_x_flow_cell.bcl_converter,
        flow_cell=novaseq_x_flow_cell,
        lims_samples=lims_novaseq_x_samples,
        force=False,
    )

    # THEN no error is raised and the sample sheet creator is v1
    assert isinstance(sample_sheet_creator, SampleSheetCreatorV2)



import datetime
from pathlib import Path
from typing import Dict, List

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.clean.flow_cell_run_directories import RunDirFlowCell
from cg.store import Store
from housekeeper.store.models import File, Tag, Version
from tests.store_helpers import StoreHelpers


def test_archive_flow_cell_run_directory(
    caplog,
    real_housekeeper_api: HousekeeperAPI,
    dragen_flow_cell_dir: Path,
    timestamp: datetime.datetime,
    base_store: Store,
    helpers: StoreHelpers,
):
    # GIVEN an old flow cell which is not tagged as archived but whose sample sheet is included in Housekeeper

    bundle_data: Dict = {"name": "HLG5GDRXY", "created": timestamp, "files": []}

    helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=bundle_data)

    real_housekeeper_api.add_and_include_file_to_latest_version(
        bundle_name="HLG5GDRXY",
        file=Path(dragen_flow_cell_dir, "SampleSheet.csv"),
        tags=["samplesheet", "HLG5GDRXY"],
    )

    # WHEN running archive_sample_sheet in the run directory
    run_dir_flow_cell: RunDirFlowCell = RunDirFlowCell(
        flow_cell_dir=dragen_flow_cell_dir,
        status_db=base_store,
        housekeeper_api=real_housekeeper_api,
    )
    run_dir_flow_cell.archive_sample_sheet()

    # THEN assert the archiving is skipped
    assert "Sample sheet already included!" in caplog.text


def test_filter_sample_sheets(
    caplog,
    real_housekeeper_api: HousekeeperAPI,
    dragen_flow_cell_dir: Path,
    timestamp: datetime.datetime,
    base_store: Store,
    helpers: StoreHelpers,
):
    archived_sample_sheet: Tag = Tag(id=1, name="archived_sample_sheet")
    samplesheet: Tag = Tag(id=2, name="samplesheet")
    # GIVEN a list of files
    file1: File = File(id=1, path="path/to/file1", tags=[])
    file2: File = File(id=2, path="path/to/file2", tags=[archived_sample_sheet])
    file3: File = File(id=3, path="path/to/file3", tags=[samplesheet])
    file4: File = File(id=4, path="path/to/file4", tags=[archived_sample_sheet, samplesheet])
    version: Version = Version(id=1, bundle_id=1, files=[file1, file2, file3, file4])

    bundle_data: Dict = {"name": "HLG5GDRXY", "created": timestamp, "files": [], version: version}
    helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=bundle_data)

    # WHEN filtering out files tagged with neither archived_sample_sheet nor samplesheet
    run_dir_flow_cell: RunDirFlowCell = RunDirFlowCell(
        flow_cell_dir=dragen_flow_cell_dir,
        status_db=base_store,
        housekeeper_api=real_housekeeper_api,
    )
    filtered_files: List[File] = run_dir_flow_cell.filter_on_sample_sheets(
        [file1, file2, file3, file4]
    )

    # THEN the list should contain file2, file3 and file4 but not file1
    assert file1 not in filtered_files
    assert file2 in filtered_files
    assert file3 in filtered_files
    assert file4 in filtered_files



"""
This module handles concatenation of balsamic fastq files.

Classes:
    FastqFileNameCreator: Creates valid balsamic filenames
    FastqHandler: Handles fastq file linking
"""
import datetime as dt
import gzip
import logging
import os
import re
import shutil
from pathlib import Path
from typing import List, Optional


LOG = logging.getLogger(__name__)

DEFAULT_DATE_STR = (
    "171015"  # Stand in value to use if sequencing date cannot be extracted from header
)
DEFAULT_INDEX = (
    "XXXXXX"  # Stand in value to use if flowcell index is to be masked when renaming file
)


class FastqHandler:
    """Handles fastq file linking"""

    @staticmethod
    def concatenate(files: List, concat_file: str):
        """Concatenates a list of fastq files"""
        LOG.info(FastqHandler.display_files(files, concat_file))

        with open(concat_file, "wb") as write_file_obj:
            for filename in files:
                with open(filename, "rb") as file_descriptor:
                    shutil.copyfileobj(file_descriptor, write_file_obj)

        size_before = FastqHandler.size_before(files)
        size_after = FastqHandler.size_after(concat_file)

        try:
            FastqHandler.assert_file_sizes(size_before, size_after)
        except AssertionError as error:
            LOG.warning(error)

    @staticmethod
    def size_before(files: List) -> int:
        """Returns the total size of the linked fastq files before concatenation"""

        return sum(os.stat(f).st_size for f in files)

    @staticmethod
    def size_after(concat_file: str) -> int:
        """returns the size of the concatenated fastq files"""

        return os.stat(concat_file).st_size

    @staticmethod
    def assert_file_sizes(size_before: int, size_after: int) -> None:
        """asserts the file sizes before and after concatenation. If the file sizes differ by more
        than 1 percent, throw an exception"""
        msg = (
            f"Warning: Large file size difference after concatenation!"
            f"Before: {size_before} -> after: {size_after}"
        )

        assert abs(size_before - size_after) / size_before <= 0.01, msg
        LOG.info("Concatenation file size check successful!")

    @staticmethod
    def display_files(files: List, concat_file: str) -> str:
        """display file names for logging purposes"""
        return (
            f"Concatenating: {', '.join(Path(file_).name for file_ in files)} -> "
            f"{Path(concat_file).name}"
        )

    @staticmethod
    def remove_files(files: list) -> None:
        for file_name in files:
            file_path: Path = Path(file_name)
            if not file_path.is_file():
                continue
            file_path.unlink()

    @staticmethod
    def get_concatenated_name(linked_fastq_name: str) -> str:
        """ "create a name for the concatenated file for some read files"""
        return f"concatenated_{'_'.join(linked_fastq_name.split('_')[-4:])}"

    @staticmethod
    def parse_header(line: str) -> dict:
        """Generates a dict with parsed lanes, flowcells and read numbers
        Handle illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format

        @HWUSI-EAS100R:6:73:941:1973#0/1

            HWUSI-EAS100R   the unique instrument name
            6   flowcell lane
            73  tile number within the flowcell lane
            941     'x'-coordinate of the cluster within the tile
            1973    'y'-coordinate of the cluster within the tile
            #0  index number for a multiplexed sample (0 for no indexing)
            /1  the member of a pair, /1 or /2 (paired-end or mate-pair reads only)

        Versions of the Illumina pipeline since 1.4 appear to use #NNNNNN
        instead of #0 for the multiplex ID, where NNNNNN is the sequence of the
        multiplex tag.

        With Casava 1.8 the format of the '@' line has changed:

        @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG

            EAS139  the unique instrument name
            136     the run id
            FC706VJ     the flowcell id
            2   flowcell lane
            2104    tile number within the flowcell lane
            15343   'x'-coordinate of the cluster within the tile
            197393  'y'-coordinate of the cluster within the tile
            1   the member of a pair, 1 or 2 (paired-end or mate-pair reads only)
            Y   Y if the read is filtered, N otherwise
            18  0 when none of the control bits are on, otherwise it is an even number
            ATCACG  index sequence
        """

        fastq_meta = {"lane": None, "flowcell": None, "readnumber": None}

        parts = line.split(":")
        if len(parts) == 5:  # @HWUSI-EAS100R:6:73:941:1973#0/1
            fastq_meta["lane"] = parts[1]
            fastq_meta["flowcell"] = "XXXXXX"
            fastq_meta["readnumber"] = parts[-1].split("/")[-1]
        if len(parts) == 10:  # @EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG
            fastq_meta["lane"] = parts[3]
            fastq_meta["flowcell"] = parts[2]
            fastq_meta["readnumber"] = parts[6].split(" ")[-1]
        if len(parts) == 7:  # @ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1
            fastq_meta["lane"] = parts[3]
            fastq_meta["flowcell"] = parts[2]
            fastq_meta["readnumber"] = parts[-1].split("/")[-1]

        return fastq_meta

    @staticmethod
    def parse_file_data(fastq_path: Path) -> dict:
        with gzip.open(fastq_path) as handle:
            header_line = handle.readline().decode()
            header_info = FastqHandler.parse_header(header_line)

            data = {
                "path": fastq_path,
                "lane": int(header_info["lane"]),
                "flowcell": header_info["flowcell"],
                "read": int(header_info["readnumber"]),
                "undetermined": ("Undetermined" in fastq_path),
            }
            matches = re.findall(r"-l[1-9]t([1-9]{2})_", str(fastq_path))
            if len(matches) > 0:
                data["flowcell"] = f"{data['flowcell']}-{matches[0]}"
            return data

    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ):
        raise NotImplementedError


class BalsamicFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following Balsamic conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_R_{read}.fastq.gz"


class MipFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions."""
        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        date_str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date_str}_{flowcell}_{sample}_{index}_{read}.fastq.gz"


class MicrosaltFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following usalt conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz

        flowcell = f"{flowcell}-undetermined" if undetermined else flowcell
        return f"{sample}_{flowcell}_L{lane}_{read}.fastq.gz"


class MutantFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flowcell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following mutant conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        # ACC1234A1_FCAB1ABC2_L1_1.fastq.gz sample_flowcell_lane_read.fastq.gz

        return f"{flowcell}_L{lane}_{meta}_{read}.fastq.gz"

    @staticmethod
    def get_concatenated_name(linked_fastq_name: str) -> str:
        """ "Create a name for the concatenated file from multiple lanes"""
        return "_".join(linked_fastq_name.split("_")[2:])

    @staticmethod
    def get_nanopore_header_info(line: str) -> dict:
        fastq_meta = {"flowcell": None}
        header_metadata: list = line.split(" ")
        flowcell = header_metadata[5].split("=")
        fastq_meta["flowcell"] = flowcell[1]
        return fastq_meta

    @staticmethod
    def create_nanopore_fastq_name(
        flowcell: str,
        sample: str,
        filenr: str,
        meta: Optional[str] = None,
    ) -> str:
        return f"{flowcell}_{sample}_{meta}_{filenr}.fastq.gz"

    @staticmethod
    def parse_nanopore_file_data(fastq_path: Path) -> dict:
        with gzip.open(fastq_path) as handle:
            header_line = handle.readline().decode()
            header_info: dict = MutantFastqHandler.get_nanopore_header_info(line=header_line)
            data = {
                "path": fastq_path,
                "flowcell": header_info["flowcell"],
            }
            return data


class RnafusionFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: str,
        flow_cell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions,
        no naming constrains from pipeline."""
        flow_cell: str = f"{flow_cell}-undetermined" if undetermined else flow_cell
        date: str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date}_{flow_cell}_{sample}_{index}_{read}.fastq.gz"


class TaxprofilerFastqHandler(FastqHandler):
    """Handles Taxprofiler fastq file linking."""

    @staticmethod
    def create_fastq_name(
        lane: str,
        flow_cell: str,
        sample: str,
        read: str,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: Optional[str] = None,
        meta: Optional[str] = None,
    ) -> str:
        """Name a FASTQ file, no naming constrains from pipeline."""
        flow_cell: str = f"{flow_cell}-undetermined" if undetermined else flow_cell
        date: str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date}_{flow_cell}_{sample}_{index}_{read}.fastq.gz"



"""Rnafusion quality control metrics model."""
from typing import Optional

from pydantic.v1 import BaseModel


class RnafusionQCMetrics(BaseModel):
    """RNAfusion QC metrics."""

    after_filtering_gc_content: Optional[float]
    after_filtering_q20_rate: Optional[float]
    after_filtering_q30_rate: Optional[float]
    after_filtering_read1_mean_length: Optional[float]
    before_filtering_total_reads: Optional[float]
    bias_5_3: Optional[float]
    pct_adapter: Optional[float]
    pct_mrna_bases: Optional[float]
    pct_ribosomal_bases: Optional[float]
    pct_surviving: Optional[float]
    pct_duplication: Optional[float]
    reads_aligned: Optional[float]
    uniquely_mapped_percent: Optional[float]



from typing import Dict, Optional

from pydantic.v1 import ValidationError as PydanticValidationError
import pytest

from cg.models.nextflow.deliverables import NextflowDeliverables


def test_instantiate_nextflow_deliverables(
    nextflow_deliverables: Dict[str, str],
):
    """
    Tests nextflow delivery object.
    """
    # GIVEN a sample with fastq files and strandedness

    # WHEN instantiating a deliverables object
    nextflow_deliverables_object: NextflowDeliverables = NextflowDeliverables(
        deliverables=nextflow_deliverables
    )

    # THEN assert that it was successfully created
    assert isinstance(nextflow_deliverables_object, NextflowDeliverables)


def test_instantiate_nextflow_deliverables_with_empty_entry(
    nextflow_deliverables_with_empty_entry: Dict[str, Optional[str]],
):
    """
    Tests nextflow delivery object with empty entry.
    """
    # WHEN instantiating a deliverables object with an empty entry THEN assert that it was successfully created
    with pytest.raises(PydanticValidationError):
        NextflowDeliverables(deliverables=nextflow_deliverables_with_empty_entry)


def test_instantiate_nextflow_deliverables_with_faulty_entry(
    nextflow_deliverables_with_faulty_entry: Dict[str, str],
):
    """
    Tests nextflow delivery object with empty entry.
    """
    # WHEN instantiating a deliverables object with an empty entry THEN assert that it was successfully created
    with pytest.raises(PydanticValidationError):
        NextflowDeliverables(deliverables=nextflow_deliverables_with_faulty_entry)



import logging
from typing import List
from typing_extensions import Literal

from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreator,
    SampleSheetCreatorV1,
    SampleSheetCreatorV2,
)
from cg.apps.demultiplex.sample_sheet.models import FlowCellSample
from cg.constants.sequencing import Sequencers
from cg.constants.demultiplexing import BclConverter
from cg.exc import FlowCellError
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData

LOG = logging.getLogger(__name__)


def get_sample_sheet_creator(
    bcl_converter: str,
    flow_cell: FlowCellDirectoryData,
    lims_samples: List[FlowCellSample],
    force: bool,
) -> SampleSheetCreator:
    """Returns an initialised sample sheet creator according to the flow cell sequencer."""
    sequencer: str = flow_cell.sequencer_type
    if sequencer == Sequencers.NOVASEQ:
        return SampleSheetCreatorV1(
            bcl_converter=bcl_converter, flow_cell=flow_cell, lims_samples=lims_samples, force=force
        )
    elif sequencer == Sequencers.NOVASEQX:
        return SampleSheetCreatorV2(
            bcl_converter=bcl_converter, flow_cell=flow_cell, lims_samples=lims_samples, force=force
        )
    else:
        message: str = (
            "Only NovaSeq and NovaSeqX sample sheets are currently supported."
            + f"Found sequencer type: {sequencer}"
        )
        raise FlowCellError(message)


def create_sample_sheet(
    bcl_converter: Literal[BclConverter.BCL2FASTQ, BclConverter.DRAGEN],
    flow_cell: FlowCellDirectoryData,
    lims_samples: List[FlowCellSample],
    force: bool = False,
) -> List[List[str]]:
    """Create a sample sheet for a flow cell."""
    sample_sheet_creator = get_sample_sheet_creator(
        bcl_converter=bcl_converter,
        flow_cell=flow_cell,
        lims_samples=lims_samples,
        force=force,
    )
    LOG.info(
        f"Constructing a {bcl_converter} sample sheet for the {flow_cell.sequencer_type} flow cell {flow_cell.id}"
    )
    return sample_sheet_creator.construct_sample_sheet()



"""Taxprofiler related constants."""


from cg.constants.nextflow import NFX_SAMPLE_HEADER, NFX_SAMPLESHEET_READS_HEADERS
from typing import List

TAXPROFILER_RUN_ACCESSION: str = "run_accession"
TAXPROFILER_INSTRUMENT_PLATFORM: str = "instrument_platform"
TAXPROFILER_FASTA_HEADER: str = "fasta"
TAXPROFILER_SAMPLE_SHEET_HEADERS: List[str] = [
    NFX_SAMPLE_HEADER,
    TAXPROFILER_RUN_ACCESSION,
    TAXPROFILER_INSTRUMENT_PLATFORM,
]

TAXPROFILER_SAMPLE_SHEET_HEADERS += NFX_SAMPLESHEET_READS_HEADERS
TAXPROFILER_SAMPLE_SHEET_HEADERS.append(TAXPROFILER_FASTA_HEADER)



"""Tests cli methods to create the case config for Taxprofiler."""

import logging
from pathlib import Path
from typing import List

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.taxprofiler.base import config_case
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_without_options(cli_runner: CliRunner, taxprofiler_context: CGConfig):
    """Test command without options."""

    # WHEN dry running without anything specified
    result = cli_runner.invoke(config_case, obj=taxprofiler_context)

    # THEN command should mention argument
    assert result.exit_code != EXIT_SUCCESS
    assert "Missing argument" in result.output


def test_with_missing_case(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
):
    """Test command with invalid case to start with."""
    caplog.set_level(logging.ERROR)
    # GIVEN case id not in database
    assert not taxprofiler_context.status_db.get_case_by_internal_id(
        internal_id=case_id_does_not_exist
    )

    # WHEN running
    result = cli_runner.invoke(config_case, [case_id_does_not_exist], obj=taxprofiler_context)

    # THEN command should NOT successfully call the command it creates
    assert result.exit_code != EXIT_SUCCESS
    # THEN ERROR log should be printed containing invalid case_id
    assert "could not be found in StatusDB!" in caplog.text



from pathlib import Path
from typing import Optional, Union

from pydantic.v1 import BaseModel


class CommandArgs(BaseModel):
    """Model for arguments and options supported."""

    log: Optional[Union[str, Path]]
    resume: Optional[bool]
    profile: Optional[str]
    stub: Optional[bool]
    config: Optional[Union[str, Path]]
    name: Optional[str]
    revision: Optional[str]
    wait: Optional[str]
    id: Optional[str]
    with_tower: Optional[bool]
    use_nextflow: Optional[bool]
    compute_env: Optional[str]
    work_dir: Optional[Union[str, Path]]
    params_file: Optional[Union[str, Path]]



from pathlib import Path
from typing import Type

from cg.apps.demultiplex.sample_sheet.models import (
    FlowCellSampleNovaSeq6000Bcl2Fastq,
    FlowCellSampleNovaSeq6000Dragen,
    FlowCellSampleNovaSeqX,
)
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


def test_flowcell_id(bcl2fastq_flow_cell_dir: Path):
    """Test parsing of flow cell id."""
    # GIVEN the path to a finished flow cell run
    # GIVEN the flow cell id
    flowcell_id: str = bcl2fastq_flow_cell_dir.name.split("_")[-1][1:]

    # WHEN instantiating a flow cell object
    flowcell_obj = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)

    # THEN assert that the flow cell id is parsed
    assert flowcell_obj.id == flowcell_id


def test_flowcell_position(bcl2fastq_flow_cell_dir: Path):
    """Test getting flow cell position."""
    # GIVEN the path to a finished flow cell
    # GIVEN a flow cell object
    flowcell_obj = FlowCellDirectoryData(flow_cell_path=bcl2fastq_flow_cell_dir)

    # WHEN fetching the flow cell position
    position = flowcell_obj.position

    # THEN assert it is A or B
    assert position in ["A", "B"]


def test_rta_exists(bcl2fastq_flow_cell: FlowCellDirectoryData):
    """Test return of RTS file."""
    # GIVEN the path to a finished flow cell
    # GIVEN a flow cell object

    # WHEN fetching the path to the RTA file
    rta_file: Path = bcl2fastq_flow_cell.rta_complete_path

    # THEN assert that the file exists
    assert rta_file.exists()


def test_is_hiseq_x_copy_completed_ready(bcl2fastq_flow_cell: FlowCellDirectoryData):
    # GIVEN the path to a demultiplexed finished flow cell
    # GIVEN a flow cell object
    # GIVEN a copy complete file

    # WHEN fetching the path to the copy complete file
    is_completed = bcl2fastq_flow_cell.is_hiseq_x_copy_completed()

    # THEN assert that the file exists
    assert is_completed is True


def test_is_hiseq_x_delivery_started_ready(
    bcl2fastq_flow_cell: FlowCellDirectoryData, demultiplexing_delivery_file: Path
):
    # GIVEN the path to a demultiplexed finished flow cell
    # GIVEN a flow cell object
    # GIVEN a delivery file
    demultiplexing_delivery_file.touch()

    # WHEN checking the path to the delivery file
    is_delivered = bcl2fastq_flow_cell.is_hiseq_x_delivery_started()

    demultiplexing_delivery_file.unlink()

    # THEN assert that the file exists
    assert is_delivered is True


def test_is_hiseq_x_delivery_started_not_ready(bcl2fastq_flow_cell: FlowCellDirectoryData):
    # GIVEN the path to a demultiplexed finished flow cell
    # GIVEN a flow cell object

    # WHEN checking the path to the copy complete file
    is_delivered = bcl2fastq_flow_cell.is_hiseq_x_delivery_started()

    # THEN assert that the file do not exist
    assert is_delivered is False


def test_is_hiseq_x(bcl2fastq_flow_cell: FlowCellDirectoryData, hiseq_x_tile_dir: Path):
    # GIVEN the path to a demultiplexed finished flow cell
    # GIVEN a flow cell object
    # GIVEN a Hiseq X directory
    hiseq_x_tile_dir.mkdir(exist_ok=True)

    # WHEN checking the path to the Hiseq X flow cell directory
    is_hiseq_x = bcl2fastq_flow_cell.is_hiseq_x()

    # Clean up
    hiseq_x_tile_dir.rmdir()

    # THEN assert that the file exists
    assert is_hiseq_x is True


def test_get_sample_model_bcl2fastq(bcl2fastq_flow_cell: FlowCellDirectoryData):
    """Test that the sample model of a bcl2fastq flow cell is FlowCellSampleNovaSeq6000Bcl2Fastq."""
    # GIVEN a Bcl2Fastq flow cell

    # WHEN getting the sample model
    sample_model: Type[FlowCellSampleNovaSeq6000Bcl2Fastq] = bcl2fastq_flow_cell.sample_type

    # THEN it is FlowCellSampleNovaSeq6000Bcl2Fastq
    assert sample_model == FlowCellSampleNovaSeq6000Bcl2Fastq


def test_get_sample_model_dragen(dragen_flow_cell: FlowCellDirectoryData):
    """Test that the sample model of a dragen flow cell is FlowCellSampleNovaSeq6000Dragen."""
    # GIVEN a dragen flow cell

    # WHEN getting the sample model
    sample_model: Type[FlowCellSampleNovaSeq6000Dragen] = dragen_flow_cell.sample_type

    # THEN it is FlowCellSampleNovaSeq6000Bcl2Fastq
    assert sample_model == FlowCellSampleNovaSeq6000Dragen


def test_get_sample_model_novaseq_x(novaseq_x_flow_cell: FlowCellDirectoryData):
    """Test that the sample model of a NovaSeqX flow cell is FlowCellSampleNovaSeqX."""
    # GIVEN a NovaSeqX flow cell

    # WHEN getting the sample model
    sample_model: Type[FlowCellSampleNovaSeqX] = novaseq_x_flow_cell.sample_type

    # THEN it is FlowCellSampleNovaSeq6000Bcl2Fastq
    assert sample_model == FlowCellSampleNovaSeqX



import logging
from pathlib import Path
from typing import Generator, List
from mock import MagicMock, call

from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.meta.demultiplex.demux_post_processing import (
    DemuxPostProcessingAPI,
    DemuxPostProcessingHiseqXAPI,
)
from cg.meta.transfer import TransferFlowCell
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData
from cg.store import Store


def test_set_dry_run(
    demultiplex_context: CGConfig,
):
    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingAPI = DemuxPostProcessingAPI(config=demultiplex_context)

    # THEN dry run should be False
    assert post_demux_api.dry_run is False

    # WHEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # THEN dry run should be True
    assert post_demux_api.dry_run is True


def test_add_to_cgstats_dry_run(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # When adding to cgstats
    post_demux_api.add_to_cgstats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should just log and exit
    assert "Dry run will not add flow cell stats" in caplog.text


def test_add_to_cgstats(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # When adding to cgstats
    post_demux_api.add_to_cgstats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should run the command
    assert f"add --machine X -u Unaligned {bcl2fastq_flow_cell.path}" in caplog.text


def test_cgstats_select_project_dry_run(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    cgstats_select_project_log_file: Path,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN an unaligned project directory
    Path(bcl2fastq_flow_cell.path, "Unaligned", f"Project_{flow_cell_project_id}").mkdir(
        parents=True, exist_ok=True
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # When processing project with cgstats
    post_demux_api.cgstats_select_project(
        flow_cell_id=bcl2fastq_flow_cell.id, flow_cell_path=bcl2fastq_flow_cell.path
    )

    # THEN we should just log and exit
    assert "Dry run will not process selected project" in caplog.text


def test_cgstats_select_project(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    cgstats_select_project_log_file: Path,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN an unaligned project directory
    Path(bcl2fastq_flow_cell.path, "Unaligned", f"Project_{flow_cell_project_id}").mkdir(
        parents=True, exist_ok=True
    )

    # When processing project with cgstats
    post_demux_api.cgstats_select_project(
        flow_cell_id=bcl2fastq_flow_cell.id, flow_cell_path=bcl2fastq_flow_cell.path
    )

    # THEN we should have created a stats outfile
    assert cgstats_select_project_log_file.exists()

    # Clean up from calling cgstats_select_project
    cgstats_select_project_log_file.unlink()

    # THEN we should run the command
    assert f"select --project {flow_cell_project_id} {bcl2fastq_flow_cell.id}" in caplog.text


def test_cgstats_lanestats_dry_run(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # When processing lane stats with cgstats
    post_demux_api.cgstats_lanestats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should run the command
    assert "Dry run will not add lane stats" in caplog.text


def test_cgstats_lanestats(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # When processing lane stats with cgstats
    post_demux_api.cgstats_lanestats(flow_cell_path=bcl2fastq_flow_cell.path)

    # THEN we should run the command
    assert f"lanestats {bcl2fastq_flow_cell.path}" in caplog.text


def test_finish_flow_cell_copy_not_completed(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_copy_complete_file: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a not completely copied flow cell
    if hiseq_x_copy_complete_file.exists():
        hiseq_x_copy_complete_file.unlink()

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # Reinstate
    hiseq_x_copy_complete_file.touch()

    # THEN we should log that copy is not complete
    assert f"{bcl2fastq_flow_cell.full_name} is not yet completely copied" in caplog.text


def test_finish_flow_cell_delivery_started(
    caplog,
    demultiplexing_delivery_file: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN an already started flag file
    demultiplexing_delivery_file.touch()

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # Clean up
    demultiplexing_delivery_file.unlink()

    # THEN we should log that the delivery has already started
    assert (
        f"{bcl2fastq_flow_cell.full_name} copy is complete and delivery has already started"
        in caplog.text
    )


def test_finish_flow_cell_delivery_not_hiseq_x(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_tile_dir: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN no hiseq X flow cell
    if hiseq_x_tile_dir.exists():
        hiseq_x_tile_dir.rmdir()

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # THEN we should log that this is not an Hiseq X flow cell
    assert f"{bcl2fastq_flow_cell.full_name} is not an Hiseq X flow cell" in caplog.text


def test_finish_flow_cell_ready(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    demultiplexing_delivery_file: Path,
    hiseq_x_tile_dir: Path,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    caplog.set_level(logging.INFO)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # GIVEN a Hiseq X tile directory
    hiseq_x_tile_dir.mkdir(parents=True, exist_ok=True)

    # GIVEN an unaligned project directory
    Path(
        bcl2fastq_flow_cell.path,
        DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        f"Project_{flow_cell_project_id}",
    ).mkdir(parents=True, exist_ok=True)

    # WHEN finishing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # THEN we should log that post-processing will begin
    assert (
        f"{bcl2fastq_flow_cell.full_name} copy is complete and delivery will start" in caplog.text
    )


def test_post_process_flow_cell_dry_run(
    bcl2fastq_demux_results: DemuxResults,
    caplog,
    demultiplexing_delivery_file: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    flowcell_store: Store,
    hiseq_x_tile_dir: Path,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a Hiseq X tile directory
    hiseq_x_tile_dir.mkdir(parents=True, exist_ok=True)

    # GIVEN an unaligned project directory
    Path(
        bcl2fastq_flow_cell.path,
        DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        f"Project_{flow_cell_project_id}",
    ).mkdir(parents=True, exist_ok=True)

    # GIVEN dry run set to True
    post_demux_api.set_dry_run(dry_run=True)

    # WHEN post-processing flow cell
    post_demux_api.post_process_flow_cell(demux_results=bcl2fastq_demux_results)

    # THEN a delivery file should not have been created
    assert not demultiplexing_delivery_file.exists()

    # THEN we should log that we will not commit
    assert "Dry run will not commit flow cell to database" in caplog.text


def test_post_process_flow_cell(
    bcl2fastq_demux_results: DemuxResults,
    caplog,
    cgstats_select_project_log_file: Path,
    demultiplexing_delivery_file: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    flow_cell_project_id: int,
    hiseq_x_tile_dir: Path,
    transfer_flow_cell_api: Generator[TransferFlowCell, None, None],
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a Hiseq X tile directory
    hiseq_x_tile_dir.mkdir(parents=True, exist_ok=True)

    # GIVEN an unaligned project directory
    Path(
        bcl2fastq_flow_cell.path,
        DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME,
        f"Project_{flow_cell_project_id}",
    ).mkdir(parents=True, exist_ok=True)

    # WHEN post-processing flow cell
    post_demux_api.post_process_flow_cell(demux_results=bcl2fastq_demux_results)

    # THEN a delivery file should have been created
    assert demultiplexing_delivery_file.exists()

    # Clean up
    cgstats_select_project_log_file.unlink()
    demultiplexing_delivery_file.unlink()

    # THEN we should also transfer the flow cell
    assert f"Flow cell added: {bcl2fastq_flow_cell.id}" in caplog.text


def test_finish_flow_cell(
    caplog,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_copy_complete_file: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a not completely copied flow cell
    hiseq_x_copy_complete_file.unlink()

    # When post-processing flow cell
    post_demux_api.finish_flow_cell(
        bcl_converter=BclConverter.BCL2FASTQ,
        flow_cell_name=bcl2fastq_flow_cell.full_name,
        flow_cell_path=bcl2fastq_flow_cell.path,
    )

    # Reinstate
    hiseq_x_copy_complete_file.touch()

    # THEN we should log that we are checking flow cell
    assert f"Check demultiplexed flow cell {bcl2fastq_flow_cell.full_name}" in caplog.text


def test_finish_all_flowcells(
    caplog,
    demultiplexed_flow_cell_working_directory: Path,
    demultiplex_context: CGConfig,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    hiseq_x_copy_complete_file: Path,
):
    caplog.set_level(logging.DEBUG)

    # GIVEN a demultiplex context

    # GIVEN a Demultiplexing post process API
    post_demux_api: DemuxPostProcessingHiseqXAPI = DemuxPostProcessingHiseqXAPI(
        config=demultiplex_context
    )

    # GIVEN a not completely copied flow cell
    hiseq_x_copy_complete_file.unlink()

    # When post-processing flow cell
    post_demux_api.finish_all_flow_cells(
        bcl_converter=BclConverter.BCL2FASTQ,
    )

    # Reinstate
    hiseq_x_copy_complete_file.touch()

    # THEN we should log that we are checking flow cell
    assert f"Check demultiplexed flow cell {bcl2fastq_flow_cell.full_name}" in caplog.text


def test_add_flow_cell_data_to_housekeeper(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    demux_post_processing_api.add_bundle_and_version_if_non_existent = MagicMock()
    demux_post_processing_api.add_tags_if_non_existent = MagicMock()
    demux_post_processing_api.add_sample_sheet_path_to_housekeeper = MagicMock()
    demux_post_processing_api.add_sample_fastq_files_to_housekeeper = MagicMock()

    flow_cell_name: str = "flow_cell_name"
    flow_cell_directory: Path = Path("some/path/to/flow/cell/directory")

    flow_cell = MagicMock()
    flow_cell.path = flow_cell_directory
    flow_cell.id = flow_cell_name

    # WHEN the flow cell data is added to housekeeper
    demux_post_processing_api.store_flow_cell_data_in_housekeeper(flow_cell)

    # THEN the bundle and version is added
    demux_post_processing_api.add_bundle_and_version_if_non_existent.assert_called_once_with(
        bundle_name=flow_cell_name
    )

    # THEN the correct tags are added
    demux_post_processing_api.add_tags_if_non_existent.assert_called_once_with(
        tag_names=[SequencingFileTag.FASTQ, SequencingFileTag.SAMPLE_SHEET, flow_cell_name]
    )

    # THEN the sample sheet is added
    demux_post_processing_api.add_sample_sheet_path_to_housekeeper.assert_called_once_with(
        flow_cell_directory=flow_cell_directory, flow_cell_name=flow_cell_name
    )

    # THEN the fastq files are added
    demux_post_processing_api.add_sample_fastq_files_to_housekeeper.assert_called_once_with(
        flow_cell
    )


def test_add_bundle_and_version_if_non_existent(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    demux_post_processing_api.hk_api.bundle = MagicMock(return_value=None)
    demux_post_processing_api.hk_api.create_new_bundle_and_version = MagicMock()

    # WHEN adding a bundle and version which does not exist
    flow_cell_name: str = "flow_cell_name"
    demux_post_processing_api.add_bundle_and_version_if_non_existent(bundle_name=flow_cell_name)

    # THEN that the expected methods were called with the expected arguments
    demux_post_processing_api.hk_api.bundle.assert_called_once_with(name=flow_cell_name)
    demux_post_processing_api.hk_api.create_new_bundle_and_version.assert_called_once_with(
        name=flow_cell_name
    )


def test_add_bundle_and_version_if_already_exists(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    mock_bundle = MagicMock()
    demux_post_processing_api.hk_api.bundle = MagicMock(return_value=mock_bundle)
    demux_post_processing_api.hk_api.create_new_bundle_and_version = MagicMock()

    # WHEN adding a bundle and version which already exists
    flow_cell_name: str = "flow_cell_name"
    demux_post_processing_api.add_bundle_and_version_if_non_existent(bundle_name=flow_cell_name)

    # THEN the bundle was retrieved
    demux_post_processing_api.hk_api.bundle.assert_called_once_with(name=flow_cell_name)

    # THEN a new bundle and version was not created
    demux_post_processing_api.hk_api.create_new_bundle_and_version.assert_not_called()


def test_add_tags_if_non_existent(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN that the tags do not exist
    demux_post_processing_api.hk_api.get_tag = MagicMock(return_value=None)
    demux_post_processing_api.hk_api.add_tag = MagicMock()

    # WHEN adding new tags
    tag_names = ["tag1", "tag2"]
    demux_post_processing_api.add_tags_if_non_existent(tag_names=tag_names)

    # THEN the expected housekeeper API methods were called to create the tags
    demux_post_processing_api.hk_api.get_tag.assert_has_calls(
        [call(name="tag1"), call(name="tag2")]
    )
    demux_post_processing_api.hk_api.add_tag.assert_has_calls(
        [call(name="tag1"), call(name="tag2")]
    )


def test_add_tags_if_all_exist(demultiplex_context: CGConfig):
    # Given a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # Mock the methods in the housekeeper API
    demux_post_processing_api.hk_api.get_tag = MagicMock(return_value=MagicMock())
    demux_post_processing_api.hk_api.add_tag = MagicMock()

    # Call the add_tags_if_non_existent method with two tag names
    tag_names = ["tag1", "tag2"]
    demux_post_processing_api.add_tags_if_non_existent(tag_names=tag_names)

    # Assert that the expected methods were called with the expected arguments
    demux_post_processing_api.hk_api.get_tag.assert_has_calls(
        [call(name="tag1"), call(name="tag2")]
    )
    demux_post_processing_api.hk_api.add_tag.assert_not_called()


def test_add_fastq_files_without_sample_id(
    demultiplex_context: CGConfig, dragen_flow_cell: FlowCellDirectoryData
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    demux_post_processing_api.add_file_to_bundle_if_non_existent = MagicMock()

    # WHEN add_fastq_files is called

    demux_post_processing_api.add_sample_fastq_files_to_housekeeper(flow_cell=dragen_flow_cell)

    # THEN add_file_if_non_existent was not called
    demux_post_processing_api.add_file_to_bundle_if_non_existent.assert_not_called()


def test_add_existing_sample_sheet(demultiplex_context: CGConfig, tmpdir_factory):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)
    demux_post_processing_api.add_file_to_bundle_if_non_existent = MagicMock()

    # GIVEN a flow cell directory and name
    flow_cell_directory: Path = Path(tmpdir_factory.mktemp("flow_cell_directory"))
    sample_sheet_file = Path(flow_cell_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
    sample_sheet_file.touch()
    flow_cell_name = "flow_cell_name"

    # WHEN a sample sheet is added
    demux_post_processing_api.add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell_directory, flow_cell_name=flow_cell_name
    )

    # THEN add_file_if_non_existent was called with expected arguments
    expected_file_path = Path(
        flow_cell_directory, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
    )
    expected_tag_names = [SequencingFileTag.SAMPLE_SHEET, flow_cell_name]

    demux_post_processing_api.add_file_to_bundle_if_non_existent.assert_called_once_with(
        file_path=expected_file_path,
        bundle_name=flow_cell_name,
        tag_names=expected_tag_names,
    )


def test_add_single_sequencing_metrics_entry_to_statusdb(
    store_with_sequencing_metrics: Store,
    demultiplex_context: CGConfig,
    flow_cell_name: str,
    sample_id: str,
    lane: int = 1,
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a sequencing metrics entry
    sequencing_metrics_entry = store_with_sequencing_metrics.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
        flow_cell_name=flow_cell_name, sample_internal_id=sample_id, lane=lane
    )

    # WHEN adding the sequencing metrics entry to the statusdb
    demux_post_processing_api.add_sequencing_metrics_to_statusdb(
        sample_lane_sequencing_metrics=[sequencing_metrics_entry]
    )

    # THEN the sequencing metrics entry was added to the statusdb
    assert demux_post_processing_api.status_db.get_metrics_entry_by_flow_cell_name_sample_internal_id_and_lane(
        flow_cell_name=flow_cell_name, sample_internal_id=sample_id, lane=lane
    )


def test_update_sample_read_count(demultiplex_context: CGConfig):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a sample id and a q30 threshold
    sample_id = "sample_1"
    q30_threshold = 0

    # GIVEN a sample and a read count
    sample = MagicMock()
    read_count = 100

    # GIVEN a mocked status_db
    status_db = MagicMock()
    status_db.get_sample_by_internal_id.return_value = sample
    status_db.get_number_of_reads_for_sample_passing_q30_threshold.return_value = read_count
    demux_post_processing_api.status_db = status_db

    # WHEN calling update_sample_read_count
    demux_post_processing_api.update_sample_read_count(sample_id, q30_threshold)

    # THEN get_sample_by_internal_id is called with the correct argument
    status_db.get_sample_by_internal_id.assert_called_with(sample_id)

    # THEN get_number_of_reads_for_sample_passing_q30_threshold is called with the correct arguments
    status_db.get_number_of_reads_for_sample_passing_q30_threshold.assert_called_with(
        sample_internal_id=sample_id,
        q30_threshold=q30_threshold,
    )

    # THEN the calculated_read_count has been updated with the read count for the sample
    assert sample.calculated_read_count == read_count


def test_post_processing_of_flow_cell_demultiplexed_with_bclconvert(
    demultiplex_context: CGConfig,
    flow_cell_directory_name_demultiplexed_with_bcl_convert: str,
    flow_cell_name_demultiplexed_with_bcl_convert: str,
    demultiplexed_flow_cells_tmp_directory: Path,
    bcl_convert_demultiplexed_flow_cell_sample_internal_ids: List[str],
    novaseq_6000_dir: Path,
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a directory with a flow cell demultiplexed with BCL Convert
    demux_post_processing_api.demux_api.out_dir = demultiplexed_flow_cells_tmp_directory
    demux_post_processing_api.demux_api.run_dir = novaseq_6000_dir

    # WHEN post processing the demultiplexed flow cell
    demux_post_processing_api.finish_flow_cell_temp(
        flow_cell_directory_name_demultiplexed_with_bcl_convert
    )

    # THEN a flow cell was created in statusdb
    assert demux_post_processing_api.status_db.get_flow_cell_by_name(
        flow_cell_name_demultiplexed_with_bcl_convert
    )

    # THEN sequencing metrics were created for the flow cell
    assert demux_post_processing_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name_demultiplexed_with_bcl_convert
    )
    # THEN the read count was calculated for all samples in the flow cell directory
    for sample_id in bcl_convert_demultiplexed_flow_cell_sample_internal_ids:
        sample = demux_post_processing_api.status_db.get_sample_by_internal_id(sample_id)
        assert sample is not None
        assert sample.calculated_read_count

    # THEN a bundle was added to Housekeeper for the flow cell
    assert demux_post_processing_api.hk_api.bundle(flow_cell_name_demultiplexed_with_bcl_convert)

    # THEN a bundle was added to Housekeeper for each sample
    for sample_id in bcl_convert_demultiplexed_flow_cell_sample_internal_ids:
        assert demux_post_processing_api.hk_api.bundle(sample_id)

    # THEN a sample sheet was added to Housekeeper
    assert demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.SAMPLE_SHEET],
        bundle=flow_cell_name_demultiplexed_with_bcl_convert,
    ).all()

    # THEN sample fastq files were added to Housekeeper tagged with FASTQ and the flow cell name
    for sample_id in bcl_convert_demultiplexed_flow_cell_sample_internal_ids:
        assert demux_post_processing_api.hk_api.get_files(
            tags=[SequencingFileTag.FASTQ, flow_cell_name_demultiplexed_with_bcl_convert],
            bundle=sample_id,
        ).all()

    # THEN a delivery file was created in the flow cell directory
    delivery_path = Path(
        demux_post_processing_api.demux_api.out_dir,
        flow_cell_directory_name_demultiplexed_with_bcl_convert,
        DemultiplexingDirsAndFiles.DELIVERY,
    )

    assert delivery_path.exists()


def test_post_processing_of_flow_cell_demultiplexed_with_bcl2fastq(
    demultiplex_context: CGConfig,
    flow_cell_directory_name_demultiplexed_with_bcl2fastq: str,
    flow_cell_name_demultiplexed_with_bcl2fastq: str,
    demultiplexed_flow_cells_tmp_directory: Path,
    hiseq_dir: Path,
    bcl2fastq_demultiplexed_flow_cell_sample_internal_ids: List[str],
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a directory with a flow cell demultiplexed with bcl2fastq
    demux_post_processing_api.demux_api.out_dir = demultiplexed_flow_cells_tmp_directory
    demux_post_processing_api.demux_api.run_dir = hiseq_dir

    # WHEN post processing the demultiplexed flow cell
    demux_post_processing_api.finish_flow_cell_temp(
        flow_cell_directory_name_demultiplexed_with_bcl2fastq
    )

    # THEN a flow cell was created in statusdb
    assert demux_post_processing_api.status_db.get_flow_cell_by_name(
        flow_cell_name_demultiplexed_with_bcl2fastq
    )

    # THEN sequencing metrics were created for the flow cell
    assert demux_post_processing_api.status_db.get_sample_lane_sequencing_metrics_by_flow_cell_name(
        flow_cell_name=flow_cell_name_demultiplexed_with_bcl2fastq
    )

    # THEN the read count was calculated for all samples in the flow cell directory
    for sample_internal_id in bcl2fastq_demultiplexed_flow_cell_sample_internal_ids:
        sample = demux_post_processing_api.status_db.get_sample_by_internal_id(sample_internal_id)
        assert sample is not None
        assert sample.calculated_read_count

    # THEN a bundle was added to Housekeeper for the flow cell
    assert demux_post_processing_api.hk_api.bundle(flow_cell_name_demultiplexed_with_bcl2fastq)

    # THEN a bundle was added to Housekeeper for each sample
    for sample_internal_id in bcl2fastq_demultiplexed_flow_cell_sample_internal_ids:
        assert demux_post_processing_api.hk_api.bundle(sample_internal_id)

    # THEN a sample sheet was added to Housekeeper
    assert demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.SAMPLE_SHEET],
        bundle=flow_cell_name_demultiplexed_with_bcl2fastq,
    ).all()

    # THEN sample fastq files were added to Housekeeper tagged with FASTQ and the flow cell name
    for sample_internal_id in bcl2fastq_demultiplexed_flow_cell_sample_internal_ids:
        assert demux_post_processing_api.hk_api.get_files(
            tags=[SequencingFileTag.FASTQ, flow_cell_name_demultiplexed_with_bcl2fastq],
            bundle=sample_internal_id,
        ).all()

    # THEN a delivery file was created in the flow cell directory
    delivery_path = Path(
        demux_post_processing_api.demux_api.out_dir,
        flow_cell_directory_name_demultiplexed_with_bcl2fastq,
        DemultiplexingDirsAndFiles.DELIVERY,
    )

    assert delivery_path.exists()


def test_add_demux_logs_to_housekeeper(
    demultiplex_context: CGConfig, dragen_flow_cell: FlowCellDirectoryData
):
    # GIVEN a DemuxPostProcessing API
    demux_post_processing_api = DemuxPostProcessingAPI(demultiplex_context)

    # GIVEN a bundle and flow cell version exists in housekeeper
    demux_post_processing_api.add_bundle_and_version_if_non_existent(
        bundle_name=dragen_flow_cell.id
    )

    # GIVEN a demux log in the run directory
    demux_log_file_paths: List[Path] = [
        Path(
            demux_post_processing_api.demux_api.run_dir,
            f"{dragen_flow_cell.full_name}",
            f"{dragen_flow_cell.id}_demultiplex.stdout",
        ),
        Path(
            demux_post_processing_api.demux_api.run_dir,
            f"{dragen_flow_cell.full_name}",
            f"{dragen_flow_cell.id}_demultiplex.stderr",
        ),
    ]
    for file_path in demux_log_file_paths:
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.touch()

    # WHEN adding the demux logs to housekeeper
    demux_post_processing_api.add_demux_logs_to_housekeeper(flow_cell=dragen_flow_cell)

    # THEN the demux log was added to housekeeper
    files = demux_post_processing_api.hk_api.get_files(
        tags=[SequencingFileTag.DEMUX_LOG],
        bundle=dragen_flow_cell.id,
    ).all()

    expected_file_names: List[str] = []
    for file_path in demux_log_file_paths:
        expected_file_names.append(file_path.name.split("/")[-1])

    # THEN the demux logs were added to housekeeper with the correct names
    assert len(files) == 2
    for file in files:
        assert file.path.split("/")[-1] in expected_file_names



"""Tests for the SampleSheetCreator classes."""
import pytest

from pathlib import Path
from typing import List
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import (
    SampleSheetCreator,
    SampleSheetCreatorV1,
    SampleSheetCreatorV2,
)
from cg.apps.demultiplex.sample_sheet.models import (
    SampleSheet,
    FlowCellSampleNovaSeq6000Bcl2Fastq,
    FlowCellSampleNovaSeq6000Dragen,
    FlowCellSampleNovaSeqX,
)
from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_validated_sample_sheet
from cg.constants.demultiplexing import BclConverter
from cg.exc import SampleSheetError
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData


def test_v2_sample_sheet_fails_with_bcl2fastq(
    novaseq_x_flow_cell: FlowCellDirectoryData, lims_novaseq_x_samples: List[FlowCellSampleNovaSeqX]
):
    """Test that creating a v2 sample sheet fails if the bcl converter is Bcl2fastq."""
    # GIVEN a NovaSeqX flow cell and samples

    # WHEN trying to instantiate a SampleSheetCreatorV2 with bcl2fastq as bcl_converter
    with pytest.raises(SampleSheetError) as exc_info:
        SampleSheetCreatorV2(
            flow_cell=novaseq_x_flow_cell,
            lims_samples=lims_novaseq_x_samples,
            bcl_converter=BclConverter.BCL2FASTQ,
        )
        # THEN an error is raised
        assert str(exc_info.value) == f"Can't use {BclConverter.BCL2FASTQ} with sample sheet v2"


def test_add_dummy_samples_for_sample_sheet_v1(
    novaseq6000_flow_cell_sample_1: FlowCellSampleNovaSeq6000Bcl2Fastq,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    """Test that dummy samples are added when needed for a NovaSeq6000 sample sheet."""
    # GIVEN a list of one NovaSeq6000 sample and a sample sheet creator with the sample
    samples: List[FlowCellSampleNovaSeq6000Bcl2Fastq] = [novaseq6000_flow_cell_sample_1]
    assert len(samples) == 1
    sample_sheet_creator = SampleSheetCreatorV1(
        flow_cell=bcl2fastq_flow_cell, lims_samples=samples, bcl_converter=BclConverter.BCL2FASTQ
    )

    # WHEN adding dummy samples
    sample_sheet_creator.add_dummy_samples()

    # THEN the list of sample has increased in size
    assert len(sample_sheet_creator.lims_samples) > 1


def test_construct_bcl2fastq_sheet(
    novaseq_bcl2fastq_sample_sheet_creator: SampleSheetCreator, project_dir: Path
):
    """Test that a created Bcl2fastq sample sheet has samples."""
    # GIVEN a V1 sample sheet creator populated with Bcl2fastq samples
    assert novaseq_bcl2fastq_sample_sheet_creator.lims_samples

    # WHEN building the sample sheet
    sample_sheet_content: List[
        List[str]
    ] = novaseq_bcl2fastq_sample_sheet_creator.construct_sample_sheet()

    # THEN a correctly formatted sample sheet was created
    sample_sheet: SampleSheet = get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=FlowCellSampleNovaSeq6000Bcl2Fastq,
    )
    assert sample_sheet.samples


def test_construct_dragen_sheet(
    novaseq_dragen_sample_sheet_creator: SampleSheetCreator, project_dir: Path
):
    """Test that a created Dragen sample sheet has samples."""
    # GIVEN a V1 sample sheet creator populated with Dragen samples
    assert novaseq_dragen_sample_sheet_creator.lims_samples

    # WHEN building the sample sheet
    sample_sheet_content: List[
        List[str]
    ] = novaseq_dragen_sample_sheet_creator.construct_sample_sheet()

    # THEN a correctly formatted sample sheet was created
    sample_sheet: SampleSheet = get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=FlowCellSampleNovaSeq6000Dragen,
    )
    assert sample_sheet.samples


def test_construct_novaseq_x_sheet(
    novaseq_x_sample_sheet_creator: SampleSheetCreator, project_dir: Path
):
    """Test that a created NovaSeqX sample sheet has samples."""
    # GIVEN a V2 sample sheet creator populated with Dragen samples
    assert novaseq_x_sample_sheet_creator.lims_samples

    # WHEN building the sample sheet
    sample_sheet_content: List[List[str]] = novaseq_x_sample_sheet_creator.construct_sample_sheet()

    # THEN a correctly formatted sample sheet was created
    sample_sheet: SampleSheet = get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=FlowCellSampleNovaSeqX,
    )
    assert sample_sheet.samples


def test_remove_unwanted_samples_dual_index(
    novaseq6000_flow_cell_sample_before_adapt_indexes: FlowCellSampleNovaSeq6000Bcl2Fastq,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
):
    """Test that a sample with dual index is not removed."""
    # GIVEN a sample sheet creator with a sample with dual index
    sample_sheet_creator: SampleSheetCreatorV1 = SampleSheetCreatorV1(
        flow_cell=bcl2fastq_flow_cell,
        lims_samples=[novaseq6000_flow_cell_sample_before_adapt_indexes],
        bcl_converter=BclConverter.BCL2FASTQ,
    )

    # WHEN removing unwanted samples
    sample_sheet_creator.remove_unwanted_samples()

    # THEN the sample is not removed
    assert len(sample_sheet_creator.lims_samples) == 1


def test_remove_unwanted_samples_no_dual_index(
    novaseq6000_flow_cell_sample_no_dual_index: FlowCellSampleNovaSeq6000Bcl2Fastq,
    bcl2fastq_flow_cell: FlowCellDirectoryData,
    caplog,
):
    """Test that samples with no dual index are removed."""
    # GIVEN a sample sheet creator with a sample without dual indexes
    sample_sheet_creator: SampleSheetCreatorV1 = SampleSheetCreatorV1(
        flow_cell=bcl2fastq_flow_cell,
        lims_samples=[novaseq6000_flow_cell_sample_no_dual_index],
        bcl_converter=BclConverter.BCL2FASTQ,
    )

    # WHEN removing unwanted samples
    sample_sheet_creator.remove_unwanted_samples()

    # THEN the only sample is removed
    assert len(sample_sheet_creator.lims_samples) == 0
    assert (
        f"Removing sample {novaseq6000_flow_cell_sample_no_dual_index} since it does not have dual index"
        in caplog.text
    )



"""cg module for cleaning databases and files."""
import logging
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Optional

import click
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scout_export import ScoutExportCase
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.tb import TrailblazerAPI
from cg.cli.workflow.commands import (
    balsamic_past_run_dirs,
    balsamic_pon_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    fluffy_past_run_dirs,
    microsalt_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
    rsync_past_run_dirs,
)
from cg.constants import FlowCellStatus
from cg.constants.constants import DRY_RUN, SKIP_CONFIRMATION
from cg.constants.housekeeper_tags import ALIGNMENT_FILE_TAGS, ScoutTag, SequencingFileTag
from cg.constants.sequencing import Sequencers
from cg.exc import FlowCellError, HousekeeperBundleVersionMissingError
from cg.meta.clean.api import CleanAPI
from cg.meta.clean.demultiplexed_flow_cells import DemultiplexedRunsFlowCell
from cg.meta.clean.flow_cell_run_directories import RunDirFlowCell
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCellDirectoryData as DemultiplexFlowCell
from cg.store import Store
from cg.store.models import Analysis, Flowcell, Sample
from cg.utils.date import get_date_days_ago, get_timedelta_from_date
from cg.utils.dispatcher import Dispatcher
from cgmodels.cg.constants import Pipeline
from housekeeper.store.models import File, Version
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Query
from tabulate import tabulate

CHECK_COLOR = {True: "green", False: "red"}
LOG = logging.getLogger(__name__)
FLOW_CELL_OUTPUT_HEADERS = [
    "Flow cell run name",
    "Flow cell id",
    "Correct name?",
    "Exists in statusdb?",
    "Fastq files in HK?",
    "Spring files in HK?",
    "Files on disk?",
    "Check passed?",
]


@click.group()
def clean():
    """Clean up processes."""
    return


for sub_cmd in [
    balsamic_past_run_dirs,
    balsamic_qc_past_run_dirs,
    balsamic_umi_past_run_dirs,
    balsamic_pon_past_run_dirs,
    fluffy_past_run_dirs,
    mip_dna_past_run_dirs,
    mip_rna_past_run_dirs,
    mutant_past_run_dirs,
    rnafusion_past_run_dirs,
    rsync_past_run_dirs,
    microsalt_past_run_dirs,
]:
    clean.add_command(sub_cmd)


@clean.command("hk-alignment-files")
@click.argument("bundle")
@DRY_RUN
@SKIP_CONFIRMATION
@click.pass_obj
def hk_alignment_files(
    context: CGConfig, bundle: str, yes: bool = False, dry_run: bool = False
) -> None:
    """Clean up alignment files in Housekeeper bundle."""
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    for tag in ALIGNMENT_FILE_TAGS:
        tag_files = set(housekeeper_api.get_files(bundle=bundle, tags=[tag]))

        if not tag_files:
            LOG.warning(
                f"Could not find any files ready for cleaning for bundle {bundle} and tag {tag}"
            )

        for hk_file in tag_files:
            if not (yes or click.confirm(_get_confirm_question(bundle, hk_file))):
                continue

            file_path: Path = Path(hk_file.full_path)
            if hk_file.is_included and file_path.exists():
                LOG.info(f"Unlinking {file_path}")
                if not dry_run:
                    file_path.unlink()

            LOG.info(f"Deleting {file_path} from database")
            if not dry_run:
                housekeeper_api.delete_file(file_id=hk_file.id)
                housekeeper_api.commit()


@clean.command("scout-finished-cases")
@click.option(
    "--days-old",
    type=int,
    default=300,
    help="Clean alignment files with analysis dates older then given number of days",
)
@SKIP_CONFIRMATION
@DRY_RUN
@click.pass_context
def scout_finished_cases(
    context: click.Context, days_old: int, yes: bool = False, dry_run: bool = False
) -> None:
    """Clean up of solved and archived Scout cases."""
    scout_api: ScoutAPI = context.obj.scout_api
    bundles: List[str] = []
    for status in [ScoutTag.ARCHIVED.value, ScoutTag.SOLVED.value]:
        cases: List[ScoutExportCase] = scout_api.get_cases(status=status, reruns=False)
        cases_added: int = 0
        for case in cases:
            analysis_time_delta: timedelta = get_timedelta_from_date(date=case.analysis_date)
            if analysis_time_delta.days > days_old:
                bundles.append(case.id)
                cases_added += 1
        LOG.info(f"{cases_added} cases marked for alignment files removal")

    for bundle in bundles:
        context.invoke(hk_alignment_files, bundle=bundle, yes=yes, dry_run=dry_run)


@clean.command("hk-case-bundle-files")
@click.option(
    "--days-old",
    type=int,
    default=365,
    help="Clean all files with analysis dates older then given number of days",
)
@DRY_RUN
@click.pass_context
def hk_case_bundle_files(context: CGConfig, days_old: int, dry_run: bool = False) -> None:
    """Clean up all non-protected files for all pipelines."""
    housekeeper_api: HousekeeperAPI = context.obj.housekeeper_api
    clean_api: CleanAPI = CleanAPI(status_db=context.obj.status_db, housekeeper_api=housekeeper_api)

    size_cleaned: int = 0
    version_file: File
    for version_file in clean_api.get_unprotected_existing_bundle_files(
        before=get_date_days_ago(days_ago=days_old)
    ):
        file_path: Path = Path(version_file.full_path)
        file_size: int = file_path.stat().st_size
        size_cleaned += file_size
        if dry_run:
            LOG.info(f"Dry run: {dry_run}. Keeping file {file_path}")
            continue

        file_path.unlink()
        housekeeper_api.delete_file(file_id=version_file.id)
        housekeeper_api.commit()
        LOG.info(f"Removed file {file_path}. Dry run: {dry_run}")

    LOG.info(f"Process freed {round(size_cleaned * 0.0000000001, 2)} GB. Dry run: {dry_run}")


@clean.command("hk-bundle-files")
@click.option("-c", "--case-id", type=str, required=False)
@click.option("-p", "--pipeline", type=Pipeline, required=False)
@click.option("-t", "--tags", multiple=True, required=True)
@click.option("-o", "--days-old", type=int, default=30)
@DRY_RUN
@click.pass_obj
def hk_bundle_files(
    context: CGConfig,
    case_id: Optional[str],
    tags: list,
    days_old: Optional[int],
    pipeline: Optional[Pipeline],
    dry_run: bool,
):
    """Remove files found in Housekeeper bundles."""

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db

    date_threshold: datetime = get_date_days_ago(days_ago=days_old)

    function_dispatcher: Dispatcher = Dispatcher(
        functions=[
            status_db.get_analyses_started_at_before,
            status_db.get_analyses_for_case_and_pipeline_started_at_before,
            status_db.get_analyses_for_pipeline_started_at_before,
            status_db.get_analyses_for_case_started_at_before,
        ],
        input_dict={
            "case_internal_id": case_id,
            "pipeline": pipeline,
            "started_at_before": date_threshold,
        },
    )
    analyses: List[Analysis] = function_dispatcher()

    size_cleaned: int = 0
    for analysis in analyses:
        LOG.info(f"Cleaning analysis {analysis}")
        bundle_name: str = analysis.family.internal_id
        hk_bundle_version: Optional[Version] = housekeeper_api.version(
            bundle=bundle_name, date=analysis.started_at
        )
        if not hk_bundle_version:
            LOG.warning(
                f"Version not found for "
                f"bundle:{bundle_name}; "
                f"pipeline: {analysis.pipeline}; "
                f"date {analysis.started_at}"
            )
            continue

        LOG.info(
            f"Version found for "
            f"bundle:{bundle_name}; "
            f"pipeline: {analysis.pipeline}; "
            f"date {analysis.started_at}"
        )
        version_files: List[File] = housekeeper_api.get_files(
            bundle=analysis.family.internal_id, tags=tags, version=hk_bundle_version.id
        ).all()
        for version_file in version_files:
            file_path: Path = Path(version_file.full_path)
            if not file_path.exists():
                LOG.info(f"File {file_path} not on disk.")
                continue
            LOG.info(f"File {file_path} found on disk.")
            file_size = file_path.stat().st_size
            size_cleaned += file_size
            if dry_run:
                continue

            file_path.unlink()
            housekeeper_api.delete_file(version_file.id)
            housekeeper_api.commit()
            LOG.info(f"Removed file {file_path}. Dry run: {dry_run}")

    LOG.info(f"Process freed {round(size_cleaned * 0.0000000001, 2)}GB. Dry run: {dry_run}")


@clean.command("invalid-flow-cell-dirs")
@click.option("-f", "--failed-only", is_flag=True, help="Shows failed flow cells only")
@DRY_RUN
@click.pass_obj
def remove_invalid_flow_cell_directories(context: CGConfig, failed_only: bool, dry_run: bool):
    """Removes invalid flow cell directories from demultiplexed-runs"""
    status_db: Store = context.status_db
    demux_api: DemultiplexingAPI = context.demultiplex_api
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    trailblazer_api: TrailblazerAPI = context.trailblazer_api
    sample_sheets_dir: str = context.clean.flow_cells.sample_sheets_dir_name
    checked_flow_cells: List[DemultiplexedRunsFlowCell] = []
    search: str = f"%{demux_api.out_dir}%"
    fastq_files_in_housekeeper: Query = housekeeper_api.files(
        tags=[SequencingFileTag.FASTQ]
    ).filter(File.path.like(search))
    spring_files_in_housekeeper: Query = housekeeper_api.files(
        tags=[SequencingFileTag.SPRING]
    ).filter(File.path.like(search))
    for flow_cell_dir in demux_api.out_dir.iterdir():
        flow_cell: DemultiplexedRunsFlowCell = DemultiplexedRunsFlowCell(
            flow_cell_path=flow_cell_dir,
            status_db=status_db,
            housekeeper_api=housekeeper_api,
            trailblazer_api=trailblazer_api,
            sample_sheets_dir=sample_sheets_dir,
            fastq_files=fastq_files_in_housekeeper,
            spring_files=spring_files_in_housekeeper,
        )
        if not flow_cell.is_demultiplexing_ongoing_or_started_and_not_completed:
            LOG.info(f"Found flow cell ready to be checked: {flow_cell.path}!")
            checked_flow_cells.append(flow_cell)
            if not flow_cell.passed_check:
                LOG.warning(f"Invalid flow cell directory found: {flow_cell.path}")
                if dry_run:
                    continue
                LOG.warning(f"Removing {flow_cell.path}!")
                flow_cell.remove_failed_flow_cell()
        else:
            LOG.warning("Skipping check!")

    failed_flow_cells: List[DemultiplexedRunsFlowCell] = [
        flow_cell for flow_cell in checked_flow_cells if not flow_cell.passed_check
    ]
    flow_cells_to_present: List[DemultiplexedRunsFlowCell] = (
        failed_flow_cells if failed_only else checked_flow_cells
    )
    tabulate_row = [
        [
            flow_cell.run_name,
            flow_cell.id,
            flow_cell.is_correctly_named,
            flow_cell.exists_in_statusdb,
            flow_cell.fastq_files_exist_in_housekeeper,
            flow_cell.spring_files_exist_in_housekeeper,
            flow_cell.files_exist_on_disk,
            click.style(str(flow_cell.passed_check), fg=CHECK_COLOR[flow_cell.passed_check]),
        ]
        for flow_cell in flow_cells_to_present
    ]

    click.echo(
        tabulate(
            tabulate_row,
            headers=FLOW_CELL_OUTPUT_HEADERS,
            tablefmt="fancy_grid",
            missingval="N/A",
        ),
    )


@clean.command("fix-flow-cell-status")
@DRY_RUN
@click.pass_obj
def fix_flow_cell_status(context: CGConfig, dry_run: bool):
    """Set correct flow cell statuses in Statusdb."""
    status_db: Store = context.status_db
    housekeeper_api: HousekeeperAPI = context.housekeeper_api

    flow_cells_in_statusdb: List[Flowcell] = status_db.get_flow_cells_by_statuses(
        flow_cell_statuses=[FlowCellStatus.ON_DISK, FlowCellStatus.REMOVED]
    )

    LOG.info(
        f"Number of flow cells with status {FlowCellStatus.ON_DISK.value} or {FlowCellStatus.REMOVED} in Statusdb: {len(flow_cells_in_statusdb)}"
    )

    for flow_cell in flow_cells_in_statusdb:
        sample_bundle_names: List[str] = [sample.internal_id for sample in flow_cell.samples]
        are_sequencing_files_in_hk: bool = False
        are_sequencing_files_on_disk: bool = False
        try:
            are_sequencing_files_in_hk: bool = housekeeper_api.is_fastq_or_spring_in_all_bundles(
                bundle_names=sample_bundle_names
            )
            are_sequencing_files_on_disk: bool = (
                housekeeper_api.is_fastq_or_spring_on_disk_in_all_bundles(
                    bundle_names=sample_bundle_names
                )
            )
        except HousekeeperBundleVersionMissingError:
            LOG.warning(
                f"Cannot find sample bundle in Housekeeper for sample on flow cell: {flow_cell.name}"
            )
        new_status: str = (
            FlowCellStatus.ON_DISK
            if are_sequencing_files_in_hk and are_sequencing_files_on_disk
            else FlowCellStatus.REMOVED
        )

        if flow_cell.status != new_status:
            LOG.info(
                f"Setting status of flow cell {flow_cell.name} from: {flow_cell.status} to {new_status}"
            )
            if dry_run:
                continue
            flow_cell.status: str = new_status
            status_db.session.commit()


@clean.command("remove-old-flow-cell-run-dirs")
@click.option(
    "-s",
    "--sequencer",
    type=click.Choice([Sequencers.HISEQX, Sequencers.HISEQGA, Sequencers.NOVASEQ, Sequencers.ALL]),
    default="all",
    help="Specify the sequencer. Default is to remove flow cells for all sequencers",
)
@click.option(
    "-o",
    "--days-old",
    type=int,
    default=21,
    help="Specify the age in days of the flow cells to be removed.",
)
@DRY_RUN
@click.pass_obj
def remove_old_flow_cell_run_dirs(context: CGConfig, sequencer: str, days_old: int, dry_run: bool):
    """Removes flow cells from flow cell run dir based on the sequencing date and
    the sequencer type, if specified."""
    status_db: Store = context.status_db
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    if sequencer == Sequencers.ALL:
        LOG.info("Checking flow cells for all sequencers!")
        for sequencer, run_directory in context.clean.flow_cells.flow_cell_run_dirs:
            LOG.info(f"Checking directory {run_directory} of sequencer {sequencer}:")
            flow_cell_dirs: List[Path] = [
                flow_cell_dir
                for flow_cell_dir in Path(run_directory).iterdir()
                if flow_cell_dir.is_dir()
            ]
            for flow_cell_dir in flow_cell_dirs:
                clean_run_directories(
                    days_old=days_old,
                    dry_run=dry_run,
                    housekeeper_api=housekeeper_api,
                    run_directory=flow_cell_dir,
                    status_db=status_db,
                )

    else:
        run_directory: str = dict(context.clean.flow_cells.flow_cell_run_dirs).get(sequencer)
        LOG.info(f"Checking directory {run_directory} of sequencer {sequencer}:")
        flow_cell_dirs: List[Path] = [
            flow_cell_dir
            for flow_cell_dir in Path(run_directory).iterdir()
            if flow_cell_dir.is_dir()
        ]
        for flow_cell_dir in flow_cell_dirs:
            clean_run_directories(
                days_old=days_old,
                dry_run=dry_run,
                housekeeper_api=housekeeper_api,
                run_directory=flow_cell_dir,
                status_db=status_db,
            )


@clean.command("remove-old-demutliplexed-run-dirs")
@click.option(
    "-o",
    "--days-old",
    type=int,
    default=21,
    help="Specify the age in days of the flow cells to be removed.",
)
@DRY_RUN
@click.pass_obj
def remove_old_demutliplexed_run_dirs(context: CGConfig, days_old: int, dry_run: bool):
    """Removes flow cells from demultiplexed run directory."""
    status_db: Store = context.status_db
    demux_api: DemultiplexingAPI = context.demultiplex_api
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    trailblazer_api: TrailblazerAPI = context.trailblazer_api
    for flow_cell_dir in demux_api.get_all_demultiplexed_flow_cell_dirs():
        try:
            flow_cell: DemultiplexFlowCell = DemultiplexFlowCell(flow_cell_path=flow_cell_dir)
        except FlowCellError:
            continue

        if not flow_cell.is_demultiplexing_complete:
            LOG.info(
                f"Demultiplexing not finished for {flow_cell.id}. Skipping removal of the directory."
            )
            continue

        samples: List[Sample] = status_db.get_samples_from_flow_cell(flow_cell_id=flow_cell.id)
        try:
            are_sequencing_files_in_hk: bool = housekeeper_api.is_fastq_or_spring_in_all_bundles(
                bundle_names=[sample.internal_id for sample in samples]
            )
        except HousekeeperBundleVersionMissingError:
            LOG.info(
                f"No bundle found for one or more of the samples on flow cell {flow_cell.id}."
                f" Skipping removal of the directory."
            )
            continue

        demux_runs_flow_cell: DemultiplexedRunsFlowCell = DemultiplexedRunsFlowCell(
            flow_cell_path=flow_cell_dir,
            status_db=status_db,
            housekeeper_api=housekeeper_api,
            trailblazer_api=trailblazer_api,
            sample_sheets_dir=context.clean.flow_cells.sample_sheets_dir_name,
        )
        if (
            not demux_runs_flow_cell.is_demultiplexing_ongoing_or_started_and_not_completed
            and are_sequencing_files_in_hk
        ):
            LOG.info(f"Found flow cell ready to be removed: {demux_runs_flow_cell.path}!")
            if dry_run:
                continue
            LOG.warning(f"Removing {demux_runs_flow_cell.path}!")
            clean_run_directories(
                days_old=days_old,
                dry_run=dry_run,
                housekeeper_api=housekeeper_api,
                run_directory=demux_runs_flow_cell.path,
                status_db=status_db,
            )


def clean_run_directories(
    days_old: int,
    dry_run: bool,
    housekeeper_api: HousekeeperAPI,
    run_directory: Path,
    status_db: Store,
):
    """Cleans up all flow cell directories in the specified run directory."""
    LOG.info(f"Checking flow cell {run_directory.name}")
    run_dir_flow_cell: RunDirFlowCell = RunDirFlowCell(
        flow_cell_dir=run_directory, status_db=status_db, housekeeper_api=housekeeper_api
    )
    if run_dir_flow_cell.exists_in_statusdb and run_dir_flow_cell.is_retrieved_from_pdc:
        LOG.info(
            f"Skipping removal of flow cell {run_directory}, PDC retrieval status is '{run_dir_flow_cell.flow_cell_status}'!"
        )
        return
    if run_dir_flow_cell.age < days_old:
        LOG.info(
            f"Flow cell {run_directory} is {run_dir_flow_cell.age} days old and will NOT be removed."
        )
        return
    LOG.info(f"Flow cell {run_directory} is {run_dir_flow_cell.age} days old and will be removed.")
    if dry_run:
        return
    LOG.info(f"Removing flow cell run directory {run_dir_flow_cell.flow_cell_dir}.")
    run_dir_flow_cell.archive_sample_sheet()
    run_dir_flow_cell.remove_run_directory()


def _get_confirm_question(bundle, file_obj) -> str:
    """Return confirmation question."""
    return (
        f"{bundle}: remove file from file system and database: {file_obj.full_path}"
        if file_obj.is_included
        else f"{bundle}: remove file from database: {file_obj.full_path}"
    )



""" Fixtures for microsalt CLI test """


import pytest
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers


class MockLims:
    """provides a mock class overriding relevant methods for microbial cli"""

    def __init__(self):
        self.lims = self
        self.sample_id = None
        self.lims_sample = None

    def sample(self, sample_id: str):
        """return a mock sample"""

        class LimsSample:
            """A mock class for a sample coming from LIMS. It only needs a comment"""

            def __init__(self, sample_id):
                self.sample_id = sample_id
                self.sample_data = {
                    "comment": "a comment in LimsSample",
                    "project": {"id": "microbial_order_test"},
                }

            def get(self, key):
                """only here to get the sample.get('comment')"""
                return self.sample_data.get(key, "")

        # haha, it's a factory!
        if not self.lims_sample:
            self.lims_sample = LimsSample(sample_id)

        return self.lims_sample

    def get_prep_method(self, sample_id: str) -> str:
        """Return a prep method name. Needs to be in format 'dddd:dd string'"""
        self.sample_id = sample_id
        return "1337:00 Test prep method"

    def get_sequencing_method(self, sample_id: str) -> str:
        """Return a sequencing method name. Needs to be in format 'dddd:dd string'"""
        self.sample_id = sample_id
        return "1338:00 Test sequencing method"

    def get_sample_project(self, sample_id: str) -> str:
        return self.sample(sample_id).get("project").get("id")

    def get_sample_comment(self, sample_id: str) -> str:
        return self.sample(sample_id).get("comment")


class LimsFactory:
    """Just give one LIMS"""

    single_lims = None

    @classmethod
    def produce(cls):
        """Produce me one LIMS"""
        if not cls.single_lims:
            cls.single_lims = MockLims()

        return cls.single_lims


@pytest.fixture(scope="function")
def lims_api():
    """return a mocked lims"""
    return LimsFactory.produce()


@pytest.fixture(scope="function")
def base_context(cg_context: CGConfig, helpers: StoreHelpers, lims_api):
    """The click context for the microsalt cli"""
    helpers.add_microbial_sample(cg_context.status_db)
    cg_context.lims_api_ = lims_api
    cg_context.meta_apis["analysis_api"] = MicrosaltAnalysisAPI(cg_context)
    return cg_context


@pytest.fixture(name="microbial_sample_id")
def fixture_microbial_sample_id():
    """Define a name for a microbial sample"""
    return "microbial_sample_id"


@pytest.fixture(scope="function")
def invalid_ticket_number() -> str:
    return "1234560000"



"""Parse Dragen demultiplexing statistics from Demultiplex_Stats.csv """
import csv
import logging
from pathlib import Path
from typing import Dict

from pydantic.v1 import BaseModel

LOG = logging.getLogger(__name__)


class SampleConversionResults(BaseModel):
    """Class to collect the Dragen demultiplexing conversion results for a sample on a lane"""

    raw_cluster_count: int = 0
    raw_yield: int = 0
    pass_filter_cluster_count: int = 0
    pass_filter_read1_yield: int = 0
    pass_filter_read2_yield: int = 0
    pass_filter_yield: int = 0
    pass_filter_q30: int = 0
    pass_filter_read1_q30: int = 0
    pass_filter_read2_q30: int = 0
    pass_filter_quality_score_sum: int = 0
    pass_filter_quality_score: float = 0
    barcode: str = ""
    sample_id: str = ""


class DragenDemultiplexingStats:
    def __init__(self, demux_stats_path: Path):
        self.demux_stats_path = demux_stats_path
        self.parsed_stats: Dict[int, dict] = self.parse_stats_file()

    def parse_stats_file(
        self,
    ) -> Dict[int, dict]:
        """Parse the Dragen demultiplexing stats file"""
        LOG.info("Parsing Dragen demultiplexing stats file %s", self.demux_stats_path)

        parsed_stats = {}

        with open(self.demux_stats_path, mode="r") as stats_file:
            stats_reader = csv.DictReader(stats_file)
            for row in stats_reader:
                lane = int(row["Lane"])
                sample_id = row["SampleID"]
                parsed_stats[lane] = parsed_stats.get(lane, {})
                parsed_stats[lane][sample_id] = row

        return parsed_stats



from pathlib import Path

import pytest


@pytest.fixture(name="spring_file_path")
def fixture_spring_file_path() -> Path:
    """Return spring file path"""
    return Path("/path/to/spring_file.spring")


@pytest.fixture(name="backup_file_path")
def fixture_backup_file_path() -> str:
    """Return path to a file used in the backup process"""
    return "/path/to/backup_file.extension"


@pytest.fixture(name="archived_flow_cell")
def fixture_archived_flow_cell() -> Path:
    """Path of archived flow cell"""
    return Path("/path/to/archived/flow_cell.tar.gz.gpg")


@pytest.fixture(name="archived_key")
def fixture_archived_key() -> Path:
    """Path of archived key"""
    return Path("/path/to/archived/encryption_key.key.gpg")
